from os.path import exists, abspath, dirname
import time
import logging
import glob
import warnings
import itertools
import networkx as nx
import multiprocessing as mp
from scipy.special import comb

# Open Babel and RDKit libraries
from rdkit.Chem import ChemicalFeatures
from rdkit.Chem import MolFromSmiles

# Local modules
from luna.config.params import ProjectParams
from luna.mol.features import FeatureExtractor
from luna.mol.fingerprint import generate_fp_for_mols
from luna.mol.entry import Entry, MolFileEntry
from luna.mol.groups import AtomGroupPerceiver
from luna.interaction.contact import get_contacts_with
from luna.interaction.calc import InteractionCalculator
from luna.interaction.fp.shell import ShellGenerator
from luna.interaction.fp.type import IFPType
from luna.wrappers.base import MolWrapper
from luna.util import deprecated, LUNAWarning
from luna.util.default_values import *
from luna.util.exceptions import *
from luna.util.file import *
from luna.util.logging import (new_logging_file, load_default_logging_config,
                               VERBOSITY_LEVEL)
from luna.util.multiprocessing_logging import (start_mp_handler,
                                               MultiProcessingHandler)
from luna.util.jobs import ArgsGenerator, ParallelJobs

from luna.MyBio.PDB.PDBParser import PDBParser
from luna.MyBio.PDB.FTMapParser import FTMapParser
from luna.MyBio.util import download_pdb, save_to_file, get_entity_from_entry
from luna.version import __version__, has_version_compatibility

from sys import setrecursionlimit

# Set a recursion limit to avoid RecursionError with the library pickle.
setrecursionlimit(RECURSION_LIMIT)

logger = load_default_logging_config()

MAX_NPROCS = mp.cpu_count() - 1


class StructureCache:

    def __init__(self, compounds, atm_grps_mngr):
        self.compounds = compounds
        self.atm_grps_mngr = atm_grps_mngr

        self._compounds_name = set([c.full_name
                                    for c in self.compounds])

    def is_compound_cached(self, comp):
        return comp.full_name in self._compounds_name


class EntryResults:

    """Store entry results.

    Parameters
    ----------
    entry: :class:`~luna.mol.entry.Entry`
        An :class:`~luna.mol.entry.Entry` object that represents a molecule or an entire chain.
    atm_grps_mngr: :class:`~luna.mol.groups.AtomGroupsManager`
        An :class:`~luna.mol.groups.AtomGroupsManager` object that stores the perceived atoms and
        atom groups in the vicinity given by ``entry``.
    interactions_mngr: :class:`~luna.interaction.calc.InteractionsManager`
        An :class:`~luna.interaction.calc.InteractionsManager` object that interactions in the vicinity
        given by ``entry``.
    ifp: :class:`~luna.interaction.fp.fingerprint.Fingerprint`, optional
        An interaction fingerprint (IFP) generated for ``entry``.
    mfp: RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`, optional
        A molecular fingerprint generated for ``entry``.

    Attributes
    ----------
    entry : :class:`~luna.mol.entry.Entry`
    atm_grps_mngr: :class:`~luna.mol.groups.AtomGroupsManager`
    interactions_mngr: :class:`~luna.interaction.calc.InteractionsManager`
    ifp: :class:`~luna.interaction.fp.fingerprint.Fingerprint`
    mfp: RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
    version : str
        The LUNA's version with which results were generated.
    """

    def __init__(self, entry, atm_grps_mngr, interactions_mngr, ifp=None, mfp=None):
        self.entry = entry
        self.atm_grps_mngr = atm_grps_mngr
        self.interactions_mngr = interactions_mngr
        self.ifp = ifp
        self.mfp = mfp
        self.version = __version__

    def save(self, output_file, compressed=True):
        """Write the pickled representation of this object to the file ``output_file``.

        Parameters
        ----------
        output_file : str
            The output file where the pickled representation will be saved.
        compressed : bool, optional
            If True (the default), compress the pickled representation as a  gzip file (.gz).

        Raises
        -------
        FileNotCreated
            If the file could not be created.
        """
        pickle_data(self, output_file, compressed)

    @staticmethod
    def load(input_file):
        """Read the pickled representation of an `EntryResults` object from the file
        ``input_file`` and return the reconstituted object hierarchy specified therein.
        ``input_file`` can be a gzip-compressed file.

        Raises
        -------
        PKLNotReadError
            If the file could not be loaded.
        """
        return unpickle_data(input_file)


class Project:

    """Define a LUNA project.

    .. note::
        This class is not intended to be used directly because :meth:`run` is not implemented by default.
        Instead, you should use a class that inherits from `Project` and implements :meth:`run`.
        An example is the class :class:`LocalProject` that implements a custom :meth:`run` that saves results
        as local files.

    Parameters
    ----------
    entries : iterable of :class:`~luna.mol.entry.Entry`
        Entries determine the target molecule to which interactions and other properties
        will be calculated. They can be ligands, chains, etc, and can be defined in a number of ways.
        Each entry has an associated PDB file that may contain macromolecules (protein, RNA, DNA) and
        other small molecules, water, and ions. Refer to :class:`~luna.mol.entry.Entry` for more information.
    working_path : str
        Where project results will be saved.
    pdb_path : str
        Path containing local PDB files or to where the PDB files will be downloaded.
        PDB filenames must match that defined for the entries.
        If not provided, the default PDB path will be used.
    overwrite_path : bool
        If True, allow LUNA to overwrite any existing directory, which may remove files from a previous project.
        The default value is False.
    add_h : bool
        Define if you need to add hydrogens or not. The default value is True.

        .. note::
            To be cautious, it does not add hydrogens to NMR-solved structures and ligands initialized from molecular files
            (:class:`~luna.mol.entry.MolFileEntry` objects) as they usually already contain hydrogens.
    ph : float
        Control the pH and how the hydrogens are going to be added. The default value is 7.4.

        .. note::
            To be cautious, it does not modify the protonation of molecular files defined by a
            :class:`~luna.mol.entry.MolFileEntry` object.
    amend_mol : bool
        If True (the default), try to fix atomic charges, valence, and bond types for small molecules and residues at PDB files.
        Only molecules at PDB files are validated because they do not contain charge, valence, and bond types, which may
        cause molecules to be incorrectly perceived. More information :ref:`here <Ligands in PDB files>`.

        .. note::
            Molecules from external files (:class:`~luna.mol.entry.MolFileEntry` objects) will not be modified.
    mol_obj_type : {'rdkit', 'openbabel'}
        Define which library (RDKit or Open Babel) to use to parse molecules.
        The default value is 'rdkit'.
    atom_prop_file : str
        A feature definition file (FDef) containing all information needed to define a set of
        chemical or pharmacophoric features. The default value is 'LUNA.fdef', which contains
        default LUNA features definition.
    inter_calc : :class:`~luna.interaction.calc.InteractionCalculator`
        Define which and how interactions are calculated.
    binding_mode_filter : :class:`~luna.interaction.filter.BindingModeFilter`
        Define how to filter interactions based on binding modes.
    calc_mfp : bool
        If True, generate ECFP4 fingerprints for each entry in ``entries``. The default value is False.
    mfp_output : str
        If ``calc_mfp`` is True, save ECFP4 fingerprints to file ``mfp_output``.
        If not provided, fingerprints are saved at <``working_path``>/results/fingerprints/mfp.csv.
    calc_ifp : bool
        If True (the default), generate LUNA interaction fingerprints (IFPs) for each entry in ``entries``.
    ifp_num_levels : int
        The maximum number of iterations for fingerprint generation. The default value is 2.
    ifp_radius_step : float
        The multiplier used to increase shell size at each iteration.
        At iteration 0, shell radius is 0 * ``radius_step``, at iteration 1, radius is
        1 * ``radius_step``, etc. The default value is 5.73171.
    ifp_length : int
        The fingerprint length (total number of bits). The default value is 4096.
    ifp_count : bool
        If True (the default), create a count fingerprint (:class:`~luna.interaction.fp.fingerprint.CountFingerprint`).
        Otherwise, return a bit fingerprint (:class:`~luna.interaction.fp.fingerprint.Fingerprint`).
    ifp_diff_comp_classes :
        If True (the default), include differentiation between compound classes.
        That means structural information originated from :class:`~luna.mol.groups.AtomGroup` objects
        belonging to residues, nucleotides,  ligands, or water molecules will be considered different
        even if their structural information are the same.
        This is useful for example to differentiate protein-ligand interactions from residue-residue ones.
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
        The fingerprint type (EIFP, FIFP, or HIFP). The default value is EIFP.
    ifp_output : str
        If ``calc_ifp`` is True, save LUNA interaction fingerprints (IFPs) to file ``ifp_output``.
        If not provided, fingerprints are saved at <``working_path``>/results/fingerprints/ifp.csv.
    ifp_sim_matrix_output : str, optional
        If provided, compute Tanimoto similarity between interaction fingerprints (IFPs)
        and save the similarity matrix to ``ifp_sim_matrix_output``.
    out_pse : bool
        If True, depict interactions save them as Pymol sessions (PSE file).
        The default value is False. PSE files are saved at <``working_path``>/results/pse.
    append_mode : bool
        If True, skip entries from processing if a result for them already exists in ``working_path``.
        This can save processing time in case additional entries are to be added to an existing project.
    verbosity : int
        Verbosity level. The higher the verbosity level the more information is displayed.
        Valid values are:

            * 4: DEBUG messages;
            * 3: INFO messages (the default);
            * 2: WARNING messages;
            * 1: ERROR messages;
            * 0: CRITICAL messages.
    logging_enabled : bool
        If True (the default), enable the logging system.
    nproc : int
        The number of CPUs to use. The default value is the ``maximum number of CPUs - 1``.
        If ``nproc`` is smaller than 1 or greater than the maximum amount of available CPUs at your PC,
        then ``nproc`` is set to its default value. If you set it to None, LUNA will be run serially.


    Attributes
    ----------
    entries : iterable of :class:`~luna.mol.entry.Entry`
    working_path : str
    pdb_path : str
    overwrite_path : bool
    add_h : bool
    ph : float
    amend_mol : bool
    mol_obj_type : {'rdkit', 'openbabel'}
    atom_prop_file : str
    inter_calc : :class:`~luna.interaction.calc.InteractionCalculator`
    binding_mode_filter : :class:`~luna.interaction.filter.BindingModeFilter`
    calc_mfp : bool
    mfp_output : str
    calc_ifp : bool
    ifp_num_levels : int
    ifp_radius_step : float
    ifp_length : int
    ifp_count : bool
    ifp_diff_comp_classes : bool
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
    ifp_output : str
    out_pse : bool
    out_ifp_sim_matrix : bool
    append_mode : bool
    logging_file : str
        The file to where logging messages are saved.
    version : str
        The LUNA's version with which results were generated.
    errors : list of tuple
        Any errors found during the processing of an entry.
        Each tuple contains the input and the exception raised during the execution of a task with that input.
    """

    def __init__(self,
                 entries,
                 working_path,
                 pdb_path=PDB_PATH,
                 overwrite_path=False,

                 add_h=True,
                 ph=7.4,
                 amend_mol=True,
                 mol_obj_type='rdkit',
                 atom_prop_file=ATOM_PROP_FILE,

                 inter_calc=None,
                 binding_mode_filter=None,

                 calc_mfp=False,
                 mfp_output=None,

                 calc_ifp=False,
                 ifp_num_levels=2,
                 ifp_radius_step=5.73171,
                 ifp_length=IFP_LENGTH,
                 ifp_count=True,
                 ifp_diff_comp_classes=True,
                 ifp_type=IFPType.EIFP,
                 ifp_output=None,
                 ifp_sim_matrix_output=None,

                 out_pse=False,
                 pse_path=None,

                 use_cache=False,
                 append_mode=False,
                 verbosity=3,
                 logging_enabled=True,
                 nproc=MAX_NPROCS):

        # Property required by self._log()
        self.logging_enabled = logging_enabled

        if mol_obj_type not in ACCEPTED_MOL_OBJ_TYPES:
            mobj_types = ["'%s'" % m for m in ACCEPTED_MOL_OBJ_TYPES]
            raise IllegalArgumentError("Invalid value for 'mol_obj_type'. "
                                       "Objects of type '%s' are not "
                                       "currently accepted. "
                                       "The available options are: %s."
                                       % (mol_obj_type,
                                          ", ".join(mobj_types)))

        self._log("info", "LUNA version: %s." % __version__)

        if (inter_calc is not None
                and isinstance(inter_calc, InteractionCalculator) is False):
            msg = ".".join([InteractionCalculator.__module__,
                            InteractionCalculator.__name__])
            raise IllegalArgumentError("The informed interaction "
                                       "configuration must be an instance "
                                       "of %s." % msg)
        elif inter_calc is None:
            self._log("info", "No interaction calculator object was defined "
                      "and the default will be used instead.")

        if append_mode:
            self._log("warning", "Append mode set ON, entries with existing "
                      "results will be skipped from the entries processing.")

        self.entries = entries
        self.working_path = working_path
        self.pdb_path = pdb_path
        self.overwrite_path = overwrite_path
        self.atom_prop_file = atom_prop_file or ATOM_PROP_FILE
        self.ph = ph
        self.amend_mol = amend_mol
        self.mol_obj_type = mol_obj_type
        self.add_h = add_h

        if inter_calc is None:
            inter_calc = InteractionCalculator(inter_config=INTERACTION_CONFIG)
        self.inter_calc = inter_calc

        self.binding_mode_filter = binding_mode_filter

        # Molecular fingerprint parameters.
        self.calc_mfp = calc_mfp
        self.mfp_output = mfp_output

        # Interaction fingerprint parameters.
        self.calc_ifp = calc_ifp
        self.ifp_num_levels = ifp_num_levels
        self.ifp_radius_step = ifp_radius_step
        self.ifp_length = ifp_length
        self.ifp_count = ifp_count
        self.ifp_diff_comp_classes = ifp_diff_comp_classes
        self.ifp_type = ifp_type
        self.ifp_output = ifp_output
        self.ifp_sim_matrix_output = ifp_sim_matrix_output

        # PSE parameters
        self.out_pse = out_pse
        self.pse_path = pse_path

        # General parameters.
        self.use_cache = use_cache
        self.append_mode = append_mode
        self.nproc = nproc

        self._loaded_logging_file = False
        self.verbosity = verbosity

        self.version = __version__

        self._paths = ["chunks", "configs", "logs", "pdbs",
                       "results/interactions", "results/fingerprints",
                       "results/pse", "results", "tmp"]
        self.errors = []

        self.cache = None

    def __call__(self):
        raise NotImplementedError("This class is not callable. Use a class "
                                  "that implements this method.")

    @property
    def project_file(self):
        """str: Where the pickled representation of the LUNA project is saved."""
        return "%s/project_v%s.pkl.gz" % (self.working_path, __version__)

    @property
    def results(self):
        """iterable of `EntryResults`: LUNA results for each entry."""
        for entry in self.entries:
            results = self.get_entry_results(entry)
            if results:
                yield results

    @property
    def interactions_mngrs(self):
        """iterable of :class:`~luna.interaction.calc.InteractionsManager`: \
            An :class:`~luna.interaction.calc.InteractionsManager` object for each entry."""
        for entry in self.entries:
            results = self.get_entry_results(entry)
            if results:
                yield results.interactions_mngr

    @property
    def atm_grps_mngrs(self):
        """iterable of :class:`~luna.mol.groups.AtomGroupsManager`: \
            An :class:`~luna.mol.groups.AtomGroupsManager` object for each entry."""
        for entry in self.entries:
            results = self.get_entry_results(entry)
            if results:
                yield results.atm_grps_mngr

    @property
    def ifps(self):
        """iterable of :class:`~luna.interaction.fp.fingerprint.Fingerprint`: \
            An interaction fingerprint (IFP) for each entry."""
        for entry in self.entries:
            results = self.get_entry_results(entry)
            if results:
                yield entry, results.ifp

    @property
    def mfps(self):
        """iterable of RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` \
            or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`: \
                A molecular fingerprint for each entry."""
        for entry in self.entries:
            results = self.get_entry_results(entry)
            if results:
                yield entry, results.mfp

    @property
    def nproc(self):
        """int: The number of CPUs to use."""
        return self._nproc

    @nproc.setter
    def nproc(self, nproc):
        if nproc is not None:
            if not isinstance(nproc, int) or isinstance(nproc, bool):
                self._log("warning", "The number of processes must be an integer value, but a(n) %s was provided instead. "
                          "Therefore, the number of processes 'nproc' was set to its maximum accepted capacity "
                          "(%d - 1 = %d)." % (nproc.__class__.__name__, mp.cpu_count(), MAX_NPROCS))
                nproc = MAX_NPROCS

            elif nproc < 1:
                self._log("warning", "It was trying to create an invalid number of processes (%s). Therefore, the number of "
                          "processes 'nproc' was set to its maximum accepted capacity (%d - 1 = %d)." % (str(nproc), mp.cpu_count(),
                                                                                                         MAX_NPROCS))
                nproc = MAX_NPROCS

            elif nproc >= mp.cpu_count():
                self._log("warning", "It was trying to create %d processes, "
                          "which is equal to or greater than the maximum "
                          "amount of available CPUs (%d). Therefore, the "
                          "number of processes 'nproc' was set to %d "
                          "to leave at least one CPU free."
                          % (nproc, mp.cpu_count(), MAX_NPROCS))
                nproc = MAX_NPROCS
        else:
            self._log("warning", "The number of processes was set to '%s'. "
                      "Therefore, LUNA will run jobs sequentially." % nproc)

        self._nproc = nproc

    @property
    def logging_enabled(self):
        """bool: If the logging system is enable or not."""
        return self._logging_enabled

    @logging_enabled.setter
    def logging_enabled(self, is_enabled):
        if not is_enabled:
            warnings.warn("Logging mode was set OFF. No logging information "
                          "will be saved from now on.", category=LUNAWarning,
                          stacklevel=2)

            logger.disabled = True
        else:
            warnings.warn("Logging mode was set ON. Logging information will "
                          "be saved from now on.", category=LUNAWarning,
                          stacklevel=2)
            logger.disabled = False

        self._logging_enabled = is_enabled

    @property
    def verbosity(self):
        """int: Verbosity level."""
        return self._verbosity

    @verbosity.setter
    def verbosity(self, verbosity):
        if verbosity not in VERBOSITY_LEVEL:
            v_levels = ", ".join(["%d (%s)"
                                  % (k, logging.getLevelName(v))
                                  for k, v in sorted(VERBOSITY_LEVEL.items())])
            error_msg = ("The informed verbosity level '%s' is not valid. "
                         "The valid levels are: %s."
                         % (repr(verbosity), v_levels))
            raise IllegalArgumentError(error_msg)
        else:
            self._log("info", "Verbosity set to: %d (%s)."
                      % (verbosity,
                         logging.getLevelName(VERBOSITY_LEVEL[verbosity])))

        self._verbosity = VERBOSITY_LEVEL[verbosity]

        # If the logging file has already been loaded, it is necessary
        # to update the logging verbosity level.
        if self._loaded_logging_file:
            self._init_logging_file(self.logging_file)

    def get_entry_results(self, entry):
        """Get results for a given entry.

        Parameters
        ----------
        entry : :class:`~luna.mol.entry.Entry`
            An entry from ``entries``.

        Returns
        -------
         : `EntryResults`
        """
        if isinstance(entry, Entry):
            entry = entry.to_string()

        pkl_file = "%s/chunks/%s.pkl.gz" % (self.working_path, entry)
        try:
            return EntryResults.load(pkl_file)
        except Exception as e:
            self._log("exception", e)

    def _log(self, level, message):
        if self.logging_enabled:
            try:
                getattr(logger, level)(message)
            except Exception:
                raise

    def _log_preferences(self):
        self._log("debug", "New project initialized...")
        params = []
        for key in sorted(self.__dict__):
            if key == "entries":
                params.append("\t\t\t-- # %s = %d"
                              % (key, len(self.__dict__[key])))
            else:
                params.append("\t\t\t-- %s = %s"
                              % (key, str(self.__dict__[key])))
        self._log("debug", "Preferences:\n%s" % "\n".join(params))

    def _init_logging_file(self, logging_filename=None, use_mp_handler=True):
        if self.logging_enabled:
            if not logging_filename:
                logging_filename = new_unique_filename(TMP_FILES)

            try:
                new_logging_file(logging_filename,
                                 logging_level=self.verbosity)

                start_mp_handler()

                self._log("info",
                          "Logging file '%s' initialized successfully."
                          % logging_filename)

                # Print preferences at the new logging file.
                self._log_preferences()

                self._loaded_logging_file = True
            except Exception as e:
                self._log("exception", e)
                raise FileNotCreated("Logging file '%s' could not be created."
                                     % logging_filename)

    def _close_logging_file(self):
        try:
            for handler in logger.handlers:
                if isinstance(handler, MultiProcessingHandler):
                    if isinstance(handler.sub_handler, logging.FileHandler):
                        handler.close()
                        logger.removeHandler(handler)
        except Exception:
            pass

    def _prepare_project_path(self, subdirs=None):
        self._log("info",
                  "Preparing project directory '%s'." % self.working_path)

        if self.pdb_path is None or not is_directory_valid(self.pdb_path):
            new_pdb_path = "%s/pdbs/" % self.working_path
            self._log("warning", "The provided PDB path '%s' is not valid "
                      "or does not exist. Therefore, PDBs will be saved at "
                      "the working path: %s" % (self.pdb_path, new_pdb_path))
            self.pdb_path = new_pdb_path

        if subdirs is None:
            subdirs = self._paths

        # Create main project directory.
        create_directory(self.working_path, self.overwrite_path)
        # Create subdirectories.
        for path in subdirs:
            create_directory("%s/%s" % (self.working_path, path))

        self._log("info", "Project directory '%s' created successfully."
                  % self.working_path)

    def _remove_empty_paths(self):
        for path in self._paths:
            remove_directory("%s/%s" % (self.working_path, path),
                             only_empty_paths=True)

    def remove_duplicate_entries(self):
        """Search and remove duplicate entries from ``entries``."""
        entries = {}
        for entry in self.entries:
            if entry.to_string() not in entries:
                entries[entry.to_string()] = entry
            else:
                self._log("debug", "An entry with id '%s' already exists in "
                          "the list of entries, so the entry %s is a "
                          "duplicate and will be removed."
                          % (entry.to_string(), entry))

        self._log("info", "The remotion of duplicate entries was finished. "
                  "%d entrie(s) were removed."
                  % (len(self.entries) - len(entries)))
        self.entries = list(entries.values())

    def _validate_entry_format(self, entry):
        if not entry.is_valid():
            raise InvalidEntry("Entry '%s' does not match a LUNA's entry "
                               "format." % entry.to_string())

    def verify_pdb_files_existence(self):
        """Verify if a local PDB file exists for each entry in ``entries``.
            If it does not find a given PDB file, then LUNA will try to download it from RCSB."""
        all_pdb_ids = set()
        to_download = set()
        for entry in self.entries:
            pdb_file = "%s/%s.pdb" % (self.pdb_path, entry.pdb_id)
            if not exists(pdb_file):
                to_download.add(entry.pdb_id)

            all_pdb_ids.add(entry.pdb_id)

        logger.info("%d PDB file(s) found at '%s' from a total of %d PDB(s). "
                    "So, %d PDB(s) need to be downloaded."
                    % ((len(all_pdb_ids) - len(to_download)), self.pdb_path,
                        len(all_pdb_ids), len(to_download)))

        if to_download:
            args = [(pdb_id, self.pdb_path) for pdb_id in to_download]
            pj = ParallelJobs(self.nproc)
            job_results = pj.run_jobs(args=args, consumer_func=download_pdb,
                                      job_name="Download PDBs")
            errors = job_results.errors

            # Warn the users for any errors found during
            # the entries processing.
            if errors:
                self._log("warning", "Number of PDBs with errors: %d. "
                          "Check the log file to see the complete list of "
                          "PDBs that failed." % len(errors))
                self._log("debug", "PDBs that failed: %s."
                          % ", ".join([e[0][0] for e in errors]))

    def _decide_hydrogen_addition(self, pdb_header, entry):
        if self.add_h:
            if "structure_method" in pdb_header:
                method = pdb_header["structure_method"]
                # If the method is not a NMR type does not add hydrogen
                # as it usually already has hydrogens.
                if method.upper() in NMR_METHODS:
                    self._log("debug", "The structure related to the entry "
                              "'%s' was obtained by NMR, so it will "
                              "not add hydrogens to it." % entry.to_string())
                    return False
            return True
        return False

    def _parse_complex(self, entry):

        pdb_file = "%s/%s.pdb" % (self.pdb_path, entry.pdb_id)
        entry.pdb_file = pdb_file

        pdb_parser = entry.parser
        if pdb_parser is None:
            pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True,
                                   FIX_EMPTY_CHAINS=True,
                                   FIX_ATOM_NAME_CONFLICT=True,
                                   FIX_OBABEL_FLAGS=False)

        if isinstance(pdb_parser, FTMapParser):
            only_compounds = [entry.get_biopython_key(full_id=True)]
            structure = pdb_parser.get_structure(entry.pdb_id,
                                                 pdb_file,
                                                 only_compounds=only_compounds)
            pdb_file = "%s/pdbs/%s.pdb" % (self.working_path,
                                           entry.to_string())
            save_to_file(structure, pdb_file)
            entry.pdb_file = pdb_file
        else:
            structure = pdb_parser.get_structure(entry.pdb_id, pdb_file)

        if isinstance(entry, MolFileEntry):
            structure = entry.get_biopython_structure(structure, pdb_parser)

        ligand = get_entity_from_entry(structure, entry)
        ligand.set_as_target(is_target=True)

        return pdb_parser, structure, ligand

    def _get_perceiver(self, add_h, cache=None):
        feats_factory_func = ChemicalFeatures.BuildFeatureFactory
        feature_factory = feats_factory_func(self.atom_prop_file)
        feature_extractor = FeatureExtractor(feature_factory)

        perceiver = AtomGroupPerceiver(feature_extractor, add_h=add_h,
                                       ph=self.ph, amend_mol=self.amend_mol,
                                       mol_obj_type=self.mol_obj_type,
                                       cache=cache,
                                       tmp_path="%s/tmp" % self.working_path)

        return perceiver

    def cache_protein_properties(self, entry):
        self._log("info",
                  "Cache memory activated. Residue information will be "
                  "stored and reutilized for all ligands.")

        pdb_parser, structure, ligand = self._parse_complex(entry)
        add_h = self._decide_hydrogen_addition(pdb_parser.get_header(), entry)

        inter_config = self.inter_calc.inter_config
        radius = inter_config.get("cache_cutoff",
                                  BOUNDARY_CONFIG["cache_cutoff"])
        nb_pairs = get_contacts_with(structure[0], ligand,
                                     level='R', radius=radius)
        nb_compounds = set([p[1] for p in nb_pairs
                            if not p[1].is_target()])

        mol_objs_dict = {}
        if isinstance(entry, MolFileEntry):
            mol_objs_dict[entry.get_biopython_key()] = entry.mol_obj

        perceiver = self._get_perceiver(add_h)
        atm_grps_mngr = perceiver.perceive_atom_groups(nb_compounds,
                                                       mol_objs_dict=mol_objs_dict)

        # Remove any edges involving the ligand.
        valid_edges = set()
        for edge in atm_grps_mngr.graph.edges:
            if any([atm.parent.is_target() for atm in edge]) is False:
                valid_edges.add(edge)
        atm_grps_mngr.graph = nx.Graph()
        atm_grps_mngr.graph.add_edges_from(valid_edges)

        # Remove dummy chain if the ligand comes from an
        # external molecular file.
        if isinstance(entry, MolFileEntry):
            chain = ligand.parent
            chain.detach_child(ligand.id)

            if len(chain.child_list) == 0:
                chain.parent.detach_child(chain.id)

        self.cache = StructureCache(nb_compounds, atm_grps_mngr)

    def _perceive_chemical_groups(self, entry, entity, ligand,
                                  add_h=False, cache=None):
        self._log("debug", "Starting pharmacophore perception "
                  "for entry '%s'" % entry.to_string())

        inter_config = self.inter_calc.inter_config
        radius = inter_config.get("bsite_cutoff",
                                  BOUNDARY_CONFIG["bsite_cutoff"])
        nb_pairs = get_contacts_with(entity, ligand, level='R', radius=radius)

        mol_objs_dict = {}
        if isinstance(entry, MolFileEntry):
            mol_objs_dict[entry.get_biopython_key()] = entry.mol_obj

        nb_compounds = set([x[1] for x in nb_pairs])

        perceiver = self._get_perceiver(add_h, cache)
        atm_grps_mngr = perceiver.perceive_atom_groups(nb_compounds,
                                                       mol_objs_dict=mol_objs_dict)

        self._log("debug", "Pharmacophore perception for entry '%s' has "
                  "finished." % entry.to_string())

        return atm_grps_mngr

    def _create_mfp(self, entry):
        if isinstance(entry, MolFileEntry):
            rdmol_lig = MolFromSmiles(MolWrapper(entry.mol_obj).to_smiles())
            rdmol_lig.SetProp("_Name", entry.mol_id)

            return generate_fp_for_mols([rdmol_lig], "morgan_fp")[0]["fp"]
        else:
            # TODO: implement support for other entries.
            self._log("warning", "Currently, it cannot generate molecular "
                      "fingerprints for instances of %s."
                      % entry.__class__.__name__)

    def _create_ifp(self, atm_grps_mngr):
        sg = ShellGenerator(self.ifp_num_levels, self.ifp_radius_step,
                            diff_comp_classes=self.ifp_diff_comp_classes,
                            ifp_type=self.ifp_type)
        sm = sg.create_shells(atm_grps_mngr)

        unique_shells = not self.ifp_count
        return sm.to_fingerprint(fold_to_length=self.ifp_length,
                                 unique_shells=unique_shells,
                                 count_fp=self.ifp_count)

    def _create_ifp_file(self):
        ifp_output = self.ifp_output or ("%s/results/fingerprints/ifp.csv"
                                         % self.working_path)
        with open(ifp_output, "w") as OUT:
            if self.ifp_count:
                OUT.write("ligand_id,on_bits,count\n")
            else:
                OUT.write("ligand_id,on_bits\n")

            for entry, ifp in self.ifps:
                if self.ifp_count:
                    fp_bits_str = "\t".join([str(idx)
                                             for idx in ifp.counts.keys()])
                    fp_count_str = "\t".join([str(count) for count
                                              in ifp.counts.values()])
                    OUT.write("%s,%s,%s\n" % (entry.to_string(), fp_bits_str,
                                              fp_count_str))
                else:
                    fp_bits_str = "\t".join([str(x) for x
                                             in ifp.get_on_bits()])
                    OUT.write("%s,%s\n" % (entry.to_string(), fp_bits_str))

    def _create_mfp_file(self):
        mfp_output = (self.mfp_output or "%s/results/fingerprints/mfp.csv"
                      % self.working_path)
        with open(mfp_output, "w") as OUT:
            OUT.write("ligand_id,on_bits\n")
            for entry, mfp in self.mfps:
                try:
                    bits = mfp.GetOnBits()
                except Exception:
                    try:
                        bits = mfp.GetNonzeroElements().keys()
                    except Exception:
                        error_msg = ("Fingerprint bits cannot be recovered "
                                     "for entry '%s'." % entry.to_string())
                        raise InvalidFingerprintType(error_msg)

                fp_str = "\t".join([str(x) for x in bits])
                OUT.write("%s,%s\n" % (entry.to_string(), fp_str))

    def _calc_similarity(self, res1, res2):
        return "%s,%s,%s" % (res1.entry.to_string(),
                             res2.entry.to_string(),
                             str(res1.ifp.calc_similarity(res2.ifp)))

    def _generate_similarity_matrix(self, output_file):
        nargs = int(comb(len(self.entries), 2))
        args = ArgsGenerator(itertools.combinations(self.results, 2), nargs)

        header = "entry1,entry2,similarity"
        pj = ParallelJobs(self.nproc)
        return pj.run_jobs(args=args, consumer_func=self._calc_similarity,
                           output_file=output_file, output_header=header,
                           job_name="Calculate similarities")

    def run(self):
        """Run LUNA. However, this method is not implemented by default.
        Instead, you should use a class that inherits from `Project` and implements :meth:`run`.
        An example is the class :class:`LocalProject` that implements a custom :meth:`run` that saves results
        as local files.
        """
        self()

    def save(self, output_file, compressed=True):
        """Write the pickled representation of this project to the file ``output_file``.

        Parameters
        ----------
        output_file : str
            The output file where the pickled representation will be saved.
        compressed : bool, optional
            If True (the default), compress the pickled representation as a  gzip file (.gz).

        Raises
        -------
        FileNotCreated
            If the file could not be created.
        """
        pickle_data(self, output_file, compressed)

    @staticmethod
    def get_project_file(working_path):
        """Get the pickled project file at ``working_path``."""
        return "%s/project_v%s.pkl.gz" % (working_path, __version__)

    @staticmethod
    def load(pathname, verbosity=3, logging_enabled=True):
        """Read the pickled representation of a `Project` object from a file or project path \
         and return the reconstituted object hierarchy specified therein. \
        The ``pathname`` can be a gzip-compressed file.

        Parameters
        ----------
        pathname : str
            A file containing the pickled representation of a `Project` object
            or the project path (``working_path``) from where the pickled representation will be recovered.
        verbosity : int
            Verbosity level. The higher the verbosity level the more information is displayed.
            Valid values are:

                * 4: DEBUG messages;
                * 3: INFO messages (the default);
                * 2: WARNING messages;
                * 1: ERROR messages;
                * 0: CRITICAL messages.
        logging_enabled : bool
            If True (the default), enable the logging system.

        Raises
        -------
        CompatibilityError
            If the project version is not compatible with the current LUNA version.
        PKLNotReadError
            If the file could not be loaded.
        IllegalArgumentError
            If the provided pathname does not exist or is an invalid file/directory.
        """

        # Check if the provided input path is a valid file
        # or a directory containing saved projects.
        if is_file_valid(pathname):
            input_file = pathname
        elif is_directory_valid(pathname):
            project_files = glob.glob("%s/project_v*.pkl.gz" % pathname)
            if len(project_files) == 1:
                input_file = project_files[0]
            elif len(project_files) == 0:
                raise PKLNotReadError("In the provided working path '%s', "
                                      "there is no saved project." % pathname)
            else:
                raise PKLNotReadError("In the provided working path '%s', "
                                      "there are multiple saved projects. "
                                      "Please, specify which one you want "
                                      "to load." % pathname)
        else:
            raise IllegalArgumentError("The provided path '%s' does not exist "
                                       "or is an invalid file/directory."
                                       % pathname)

        if not logging_enabled:
            logger.disabled = True

        logger.info("Reloading project saved in '%s'.\n" % input_file)

        proj_obj = unpickle_data(input_file)

        if has_version_compatibility(proj_obj.version):
            proj_obj._loaded_logging_file = False
            proj_obj.verbosity = verbosity
            proj_obj.logging_enabled = logging_enabled

            # Update the working path if the project has been
            # moved to a different path.
            curr_working_path = dirname(abspath(input_file))

            if proj_obj.working_path != curr_working_path:
                proj_obj.working_path = curr_working_path
                proj_obj.logging_file = ("%s/logs/project.log"
                                         % proj_obj.working_path)

            proj_obj._log("info", "Project reloaded successfully.")
            return proj_obj
        else:
            raise CompatibilityError("The project loaded from '%s' has a "
                                     "version (%s) not compatible with the "
                                     "current %s's version (%s)."
                                     % (input_file, proj_obj.version,
                                        __package__.upper(), __version__))

    @classmethod
    def from_config_file(cls, config_file=None):
        if config_file is not None and not exists(config_file):
            raise OSError("File '%s' does not exist." % config_file)

        proj_params = ProjectParams(config_file, fill_defaults=True)

        return cls(**proj_params)

    def save_config_file(self, config_file=None):
        config_file = config_file or f"{self.working_path}/configs/project.cfg"

        params = ProjectParams.from_project_obj(self)
        params.save_config_file(config_file)


class LocalProject(Project):

    """Define a local LUNA project, i.e., results are saved locally and not to a database.

        This class inherits from `Project` and implements :meth:`~luna.projects.Project.run`.

    Examples
    --------

    In this minimum example, we will calculate protein-ligand interactions for dopamine D4 complexes.

    First, we should define the ligand entries and initialize a new :class:`~luna.interaction.calc.InteractionCalculator`
    object.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.interaction.calc import InteractionCalculator
    >>> entries = list(MolFileEntry.from_file(input_file=f"{LUNA_PATH}/tutorial/inputs/MolEntries.txt",
    ...                                       pdb_id="D4", mol_file=f"{LUNA_PATH}/tutorial/inputs/ligands.mol2"))
    >>> ic = InteractionCalculator(inter_filter=InteractionFilter.new_pli_filter())

    Finally, just create the new LUNA project with desired parameters and call :meth:`~luna.projects.Project.run`.
    Here, we opted to define the parameters first as a dict, and then we pass it as an argument to `LocalProject`.

    >>> from luna import LocalProject
    >>> opts = {}
    >>> opts["working_path"] = "%s/Results/Test3" % main_path
    >>> opts["pdb_path"] = f"{LUNA_PATH}/tutorial/inputs/"
    >>> opts["entries"] = entries
    >>> opts["inter_calc"] = ic
    >>> proj_obj = LocalProject(**opts)
    >>> proj_obj.run()

    """

    def __init__(self, entries, working_path, **kwargs):
        super().__init__(entries=entries, working_path=working_path, **kwargs)

    def _process_entry(self, entry):

        start = time.time()

        self._log("debug", "Starting entry processing: %s."
                  % entry.to_string())

        try:
            # Check if the entry is in the correct format.
            # It also accepts entries whose pdb_id is defined by the filename.
            if isinstance(entry, MolFileEntry) is False:
                self._validate_entry_format(entry)

            # Entry results will be saved here.
            pkl_file = "%s/chunks/%s.pkl.gz" % (self.working_path,
                                                entry.to_string())
            if self.append_mode and exists(pkl_file):
                self._log("debug", "Since append mode is set ON, it will "
                          "skip entry '%s' because a result for this entry "
                          " already exists in the working path."
                          % entry.to_string())
                return

            pdb_parser, structure, ligand = self._parse_complex(entry)
            add_h = self._decide_hydrogen_addition(pdb_parser.get_header(),
                                                   entry)

            #
            # Perceive pharmacophoric properties
            #
            atm_grps_mngr = self._perceive_chemical_groups(entry,
                                                           structure[0],
                                                           ligand, add_h,
                                                           self.cache)
            atm_grps_mngr.entry = entry

            #
            # Calculate interactions
            #
            calc_func = self.inter_calc.calc_interactions
            interactions_mngr = calc_func(atm_grps_mngr.atm_grps)
            interactions_mngr.entry = entry

            # Create hydrophobic islands.
            atm_grps_mngr.merge_hydrophobic_atoms(interactions_mngr)

            if self.binding_mode_filter is not None:
                bm_filter_func = interactions_mngr.filter_out_by_binding_mode
                bm_filter_func(self.binding_mode_filter)

            # Generate IFP (Interaction fingerprint)
            ifp = None
            if self.calc_ifp:
                ifp = self._create_ifp(atm_grps_mngr)

            # Generate MFP (Molecular fingerprint)
            mfp = None
            if self.calc_mfp:
                mfp = self._create_mfp(entry)

            # Saving entry results.
            entry_results = EntryResults(entry, atm_grps_mngr,
                                         interactions_mngr, ifp, mfp)
            entry_results.save(pkl_file)

            # Saving interactions to CSV file.
            csv_file = ("%s/results/interactions/%s.csv"
                        % (self.working_path, entry.to_string()))
            interactions_mngr.to_csv(csv_file)

            # Saving interactions into a Pymol session.
            if self.out_pse:
                from luna.interaction.view import InteractionViewer
                pse_path = (self.pse_path
                            or "%s/results/pse/" % self.working_path)
                pse_file = "%s/%s.pse" % (pse_path, entry.to_string())
                piv = InteractionViewer(add_directional_arrows=False)
                piv.new_session([(entry, interactions_mngr,
                                  entry.pdb_file)], pse_file)

            self._log("debug",
                      "Processing of entry '%s' finished successfully."
                      % entry.to_string())

        except Exception:
            self._log("debug",
                      "Processing of entry '%s' failed. "
                      "Check the logs for more information."
                      % entry.to_string())
            raise

        proc_time = time.time() - start
        self._log("debug",
                  "Processing of entry '%s' took %.2fs."
                  % (entry.to_string(), proc_time))

    def _process_ifps(self, entry):
        start = time.time()

        self._log("debug",
                  "Starting IFP processing for entry '%s'."
                  % entry.to_string())

        try:
            pkl_file = ("%s/chunks/%s.pkl.gz"
                        % (self.working_path, entry.to_string()))

            if exists(pkl_file):
                # Reload results.
                entry_results = EntryResults.load(pkl_file)
                atm_grps_mngr = entry_results.atm_grps_mngr

                # Generate a new IFP.
                ifp = self._create_ifp(atm_grps_mngr)

                # Substitute old IFP by the new version and save the project.
                entry_results.ifp = ifp
                entry_results.save(pkl_file)
            else:
                error_msg = ("The IFP of the entry '%s' could not be "
                             "generated because its pickled data file "
                             "'%s' was not found." % (entry.to_string(),
                                                      pkl_file))
                raise FileNotFoundError(error_msg)

        except Exception:
            self._log("debug", "IFP processing for entry '%s' failed. Check "
                      "the logs for more information." % entry.to_string())
            raise

        proc_time = time.time() - start
        self._log("debug", "IFP processing for entry '%s' took %.2fs." %
                  (entry.to_string(), proc_time))

    def _process_mfps(self, entry):
        start = time.time()

        self._log("debug",
                  "Starting MFP processing for entry '%s'."
                  % entry.to_string())

        try:
            pkl_file = ("%s/chunks/%s.pkl.gz"
                        % (self.working_path, entry.to_string()))

            if exists(pkl_file):
                # Reload results.
                entry_results = EntryResults.load(pkl_file)

                # Generate a new MFP.
                mfp = self._create_mfp(entry)

                # Substitute old MFP by the new version and save the project.
                entry_results.mfp = mfp
                entry_results.save(pkl_file)
            else:
                error_msg = ("The MFP of the entry '%s' could not be "
                             "generated because its pickled data file "
                             "'%s' was not found." % (entry.to_string(),
                                                      pkl_file))
                raise FileNotFoundError(error_msg)

        except Exception:
            self._log("debug", "MFP processing for entry '%s' failed. Check "
                      "the logs for more information." % entry.to_string())
            raise

        proc_time = time.time() - start
        self._log("debug", "MFP processing for entry '%s' took %.2fs." %
                  (entry.to_string(), proc_time))

    def __call__(self):

        if self.entries is None or len(self.entries) == 0:
            warnings.warn("There is nothing to be done as no "
                          "entry was informed.", category=LUNAWarning,
                          stacklevel=2)
            return

        if self.working_path is None:
            warnings.warn("A working path must be provided.",
                          category=LUNAWarning, stacklevel=2)
            return

        start = time.time()

        self._prepare_project_path()

        self.logging_file = "%s/logs/project.log" % self.working_path
        self._init_logging_file(self.logging_file)

        self.remove_duplicate_entries()

        self.save_config_file()

        self._log("info", "It will verify the existence of PDB files and "
                  "download them as necessary.")
        self.verify_pdb_files_existence()

        self._log("info", "Entries processing will start. Number of entries "
                  "to be processed is: %d." % len(self.entries))
        self._log("info", "The number of processes was set to: %s."
                  % str(self.nproc))

        if self.use_cache:
            entry = self.entries[0]
            self.cache_protein_properties(entry)

        # Run jobs either in Parallel or Sequentially (nproc = None).
        pj = ParallelJobs(self.nproc)
        job_results = pj.run_jobs(args=self.entries,
                                  consumer_func=self._process_entry,
                                  job_name="Entries processing")
        self.errors = job_results.errors

        # Remove failed entries.
        if self.errors:
            entries_with_error = set([e[0].to_string() for e in self.errors])
            self.entries = [e for e in self.entries
                            if e.to_string() not in entries_with_error]

        # If all molecules failed, it won't try to create fingerprints.
        if len(self.entries) == 0:
            self._log("critical", "Entries processing failed.")
        else:
            self._log("info", "Entries processing finished successfully.")

            # Warn the users for any errors found
            # during the entries processing.
            if self.errors:
                self._log("warning",
                          "Number of entries with errors: %d. Check the log "
                          "file to see the complete list of entries that "
                          "failed." % len(entries_with_error))
                self._log("debug", "Entries that failed: %s."
                          % ", ".join([e for e in entries_with_error]))

            # Generate IFP/MFP files
            if self.calc_ifp:
                self._create_ifp_file()
            if self.calc_mfp:
                self._create_mfp_file()

            if self.ifp_sim_matrix_output and len(self.entries) > 1:
                self._log("info", "Calculating the Tanimoto similarity "
                          "between fingerprints.")
                self._generate_similarity_matrix(self.ifp_sim_matrix_output)

        # Save the whole project information.
        self.save(self.project_file)

        # Remove unnecessary paths.
        self._remove_empty_paths()

        end = time.time()
        self._log("info", "Project creation completed!!!")
        self._log("info", "Total processing time: %.2fs." % (end - start))
        self._log("info", "Results were saved at %s." % self.working_path)
        self._log("info", "You can reload your project from %s.\n\n"
                  % self.project_file)

        # Properly close any filehandlers.
        self._close_logging_file()

    def generate_fps(self):
        """Generate LUNA interaction fingerprints (IFPs) or
        molecular fingerprints (MFPs).

        This function can be used to generate new FPs after a project is run.
        Thus, you can reload your project, vary IFP parameters
        (``ifp_num_levels``, ``ifp_radius_step``, ``ifp_length``,
        ``ifp_count``, ``ifp_diff_comp_classes``, ``ifp_type``,
        ``ifp_output``), and call `generate_fps` to create new IFPs without
        having to run the project from the scratch.

        Examples
        --------

        In the below example, we will assume a LUNA project object
        named ``proj_obj`` already exists.

        >>> from luna.interaction.fp.type import IFPType
        >>> proj_obj.ifp_num_levels = 5
        >>> proj_obj.ifp_radius_step = 1
        >>> proj_obj.ifp_length = 4096
        >>> proj_obj.ifp_type = IFPType.EIFP
        >>> proj_obj.ifp_output = "EIFP-4096__length-5__radius-1.csv"
        >>> proj_obj.generate_ifps()
        """

        if self.entries is None or len(self.entries) == 0:
            warnings.warn("There is nothing to be done as no "
                          "entry was informed.", category=LUNAWarning,
                          stacklevel=2)
            return

        if self.working_path is None:
            warnings.warn("A working path must be provided.",
                          category=LUNAWarning, stacklevel=2)
            return

        start = time.time()

        if self.calc_ifp is False and self.calc_mfp is False:
            warnings.warn("There is nothing to be done as both "
                          "'calc_ifp' and 'calc_mfp' is set to False.",
                          category=LUNAWarning, stacklevel=2)
            return

        self.overwrite_path = False

        if self.ifp_output is None or self.mfp_output is None:
            self._prepare_project_path(subdirs=["results",
                                                "results/fingerprints"])

        # Create a new directory for logs.
        if self.logging_enabled:
            if not exists("%s/logs/" % self.working_path):
                self._prepare_project_path(subdirs=["logs"])
            self._init_logging_file(self.logging_file)

        self._log("info", "Fingerprint generation will start. "
                  "Number of entries to be processed is: %d."
                  % len(self.entries))
        self._log("info", "The number of processes was set to: %s."
                  % str(self.nproc))

        def _create_fps(args, proc_func, file_func, job_name):
            # Run jobs either in Parallel or Sequentially (nproc = None).
            pj = ParallelJobs(self.nproc)
            job_results = pj.run_jobs(args=args,
                                      consumer_func=proc_func,
                                      job_name=job_name)
            errors = job_results.errors

            tmp_entries = self.entries
            # Identify failed entries.
            if errors:
                entries_with_error = set([e[0].to_string()
                                          for e in errors])
                tmp_entries = [e for e in self.entries
                               if e.to_string() not in entries_with_error]

            success = False
            # If all molecules failed, it won't try to create the output file.
            if len(tmp_entries) == 0:
                self._log("critical", f"{job_name} failed.")
            else:
                self._log("info", f"{job_name} finished successfully.")

                success = True

                # Warn the users for any errors found during
                # the entries processing.
                if errors:
                    self._log("warning", "Number of entries with errors: %d. "
                              "Check the log file to see the complete list of "
                              "entries that failed." % len(entries_with_error))
                    self._log("debug", "Entries that failed: %s."
                              % ", ".join([e for e in entries_with_error]))

                # Create an output file by calling the provided function.
                file_func()

            return success, errors

        all_errors = []
        if self.calc_ifp:
            success, errors = _create_fps(self.entries, self._process_ifps,
                                          self._create_ifp_file,
                                          "IFPs generation")
            all_errors.extend(errors)

            if success:
                if self.ifp_sim_matrix_output and len(self.entries) > 1:
                    self._log("info", "Calculating the Tanimoto similarity "
                              "between IFPs.")
                    output_file = self.ifp_sim_matrix_output
                    self._generate_similarity_matrix(output_file)

        if self.calc_mfp:
            success, errors = _create_fps(self.entries, self._process_mfps,
                                          self._create_mfp_file,
                                          "MFPs generation")
            all_errors.extend(errors)

        self.errors = all_errors

        # Remove unnecessary paths.
        self._remove_empty_paths()

        end = time.time()
        self._log("info", "Total processing time: %.2fs." % (end - start))
        self._log("info", "Results were saved at %s.\n\n" % self.working_path)

        # Properly close any filehandlers.
        self._close_logging_file()

    @deprecated("0.12.0", msg="Use the general `generate_fps` instead.")
    def generate_ifps(self):
        if self.entries is None or len(self.entries) == 0:
            warnings.warn("There is nothing to be done as no "
                          "entry was informed.", category=LUNAWarning,
                          stacklevel=2)
            return

        if self.working_path is None:
            warnings.warn("A working path must be provided.",
                          category=LUNAWarning, stacklevel=2)
            return

        start = time.time()

        self.calc_ifp = True
        self.overwrite_path = False

        if self.ifp_output is None:
            self._prepare_project_path(subdirs=["results",
                                                "results/fingerprints"])

        # Create a new directory for logs.
        if self.logging_enabled:
            if not exists("%s/logs/" % self.working_path):
                self._prepare_project_path(subdirs=["logs"])
            self._init_logging_file(self.logging_file)

        self._log("info", "Fingerprint generation will start. "
                  "Number of entries to be processed is: %d."
                  % len(self.entries))
        self._log("info", "The number of processes was set to: %s."
                  % str(self.nproc))

        # Run jobs either in Parallel or Sequentially (nproc = None).
        pj = ParallelJobs(self.nproc)
        job_results = pj.run_jobs(args=self.entries,
                                  consumer_func=self._process_ifps,
                                  job_name="Fingerprint generation")
        self.errors = job_results.errors

        tmp_entries = self.entries
        # Remove failed entries.
        if self.errors:
            entries_with_error = set([e[0].to_string() for e in self.errors])
            tmp_entries = [e for e in self.entries
                           if e.to_string() not in entries_with_error]

        # If all molecules failed, it won't try to create the output file.
        if len(tmp_entries) == 0:
            self._log("critical", "Fingerprint generation failed.")
        else:
            self._log("info", "Fingerprint generation finished successfully.")

            # Warn the users for any errors found during
            # the entries processing.
            if self.errors:
                self._log("warning", "Number of entries with errors: %d. "
                          "Check the log file to see the complete list of "
                          "entries that failed." % len(entries_with_error))
                self._log("debug", "Entries that failed: %s."
                          % ", ".join([e for e in entries_with_error]))

            # Generate IFP/MFP files
            if self.calc_ifp:
                self._create_ifp_file()
            if self.calc_mfp:
                self._create_mfp_file()

            if self.ifp_sim_matrix_output and len(self.entries) > 1:
                self._log("info", "Calculating the Tanimoto similarity "
                          "between fingerprints.")
                self._generate_similarity_matrix(self.ifp_sim_matrix_output)

        # Remove unnecessary paths.
        self._remove_empty_paths()

        end = time.time()
        self._log("info", "Total processing time: %.2fs." % (end - start))
        self._log("info", "Results were saved at %s.\n\n" % self.working_path)

        # Properly close any filehandlers.
        self._close_logging_file()



from collections import defaultdict
from itertools import chain

import numpy as np

import networkx as nx
from networkx.algorithms.shortest_paths.weighted import single_source_dijkstra

from Bio.KDTree import KDTree

from luna.MyBio.selector import Selector, AtomSelector
from luna.MyBio.util import biopython_entity_to_mol
from luna.interaction.contact import get_proximal_compounds
from luna.interaction.type import InteractionType
from luna.mol.atom import ExtendedAtom, AtomData
from luna.mol.charge_model import OpenEyeModel
from luna.mol.features import ChemicalFeature
from luna.wrappers.base import MolWrapper
from luna.util.exceptions import MoleculeSizeError, IllegalArgumentError
from luna.util.default_values import ACCEPTED_MOL_OBJ_TYPES, COV_SEARCH_RADIUS
from luna.util import math as im
from luna.util.file import pickle_data, unpickle_data
from luna.version import __version__

import logging
logger = logging.getLogger()


SS_BOND_FEATURES = ["Atom", "Acceptor", "ChalcogenDonor", "Hydrophobic"]


class AtomGroupsManager():
    """Store and manage `AtomGroup` objects.

    Parameters
    ----------
    atm_grps : iterable of `AtomGroup`, optional
        An initial sequence of `AtomGroup` objects.
    entry : :class:`~luna.mol.entry.Entry`, optional
        The chain or molecule from where the atom groups were perceived.

    Attributes
    ----------
    entry : :class:`~luna.mol.entry.Entry`
        The chain or molecule from where the atom groups were perceived.
    graph : :py:class:`networkx.Graph`
        Represent ``entry`` as a graph and its vicinity.
    version : str
        The LUNA version when the object was created.
    """

    def __init__(self, atm_grps=None, entry=None):
        self._atm_grps = []
        self.entry = entry
        self._child_dict = {}

        self.graph = nx.Graph()

        self.version = __version__

        self._compounds = set()

        self.add_atm_grps(atm_grps)

    @property
    def atm_grps(self):
        """iterable of `AtomGroup`, read-only: The sequence of `AtomGroup` objects.\
        Additional objects should be added using the method :py:meth:`add_atm_grps`."""
        return self._atm_grps

    @property
    def compounds(self):
        """set of :class:`~luna.MyBio.PDB.Residue.Residue`, read-only:
        Compounds comprising the ``atm_grps``."""
        return self._compounds

    @property
    def child_dict(self):
        """dict, read-only: Mapping between atoms (`ExtendedAtom`) and atom groups (`AtomGroup`).

        The mapping is a dict of {tuple of `ExtendedAtom` instances : `AtomGroup`} and is automatically
        updated when :py:meth:`add_atm_grps` is called."""
        return self._child_dict

    @property
    def size(self):
        """int, read-only: The number of atom groups in ``atm_grps``."""
        return len(self._atm_grps)

    @property
    def summary(self):
        """dict, read-only: The number of physicochemical features in ``atm_grps``."""
        summary = defaultdict(int)
        for grp in self.atm_grps:
            for feature in grp.features:
                summary[feature] += 1

        return summary

    def find_atm_grp(self, atoms):
        """ Find the atom group that contains the sequence of atoms ``atoms``.

        Returns
        -------
         : `AtomGroup` or None
            An atom group object or None if ``atoms`` is not in the ``child_dict`` mapping.
        """
        return self.child_dict.get(tuple(sorted(atoms)), None)

    def get_all_interactions(self):
        """Return all interactions established by the atom groups in ``atm_grps``.

        Returns
        -------
         : set of :class:`~luna.interactions.type.InteractionType`
            All interactions.
        """
        return set(chain.from_iterable([atm_grp.interactions for atm_grp in self.atm_grps]))

    def apply_filter(self, func):
        """Apply a filtering function over the atom groups in ``atm_grps`.

        Parameters
        ----------
        func : callable
            A filtering function that returns True case an `AtomGroup` object is valid and False otherwise.

        Yields
        ------
        `AtomGroup`
            A valid `AtomGroup` object.
        """
        for atm_grp in self.atm_grps:
            if func(atm_grp):
                yield atm_grp

    def filter_by_types(self, types, must_contain_all=True):
        """Filter `AtomGroup` objects by their physicochemical features.

        Parameters
        ----------
        types : iterable of str
            A sequence of physicochemical features.
        must_contain_all : bool
            If True, an `AtomGroup` object should contain all physicochemical features in ``types`` to be accepted.
            Otherwise, it will be filtered out.

        Yields
        ------
        `AtomGroup`
            A valid `AtomGroup` object.
        """
        for atm_grp in self.atm_grps:
            if must_contain_all:
                if set(types).issubset(set(atm_grp.feature_names)):
                    yield atm_grp
            else:
                if len(set(types) & set(atm_grp.feature_names)) > 0:
                    yield atm_grp

    def add_atm_grps(self, atm_grps):
        """Add one or more `AtomGroup` objects to ``atm_grps`` and automatically update ``child_dict``."""

        atm_grps = atm_grps or []

        self._atm_grps = list(set(self._atm_grps + list(atm_grps)))

        for atm_grp in atm_grps:
            self.child_dict[tuple(sorted(atm_grp.atoms))] = atm_grp
            self._compounds.update(atm_grp.compounds)
            atm_grp.manager = self

    def remove_atm_grps(self, atm_grps):
        """Remove one or more `AtomGroup` objects from ``atm_grps`` and automatically update ``child_dict``.

        Any recursive references to the removed objects will also be cleared.
        """
        self._atm_grps = list(set(self._atm_grps) - set(atm_grps))

        for atm_grp in atm_grps:
            # ExtendedAtom objects keep a list of all AtomGroup objects to which they belong to. So, if we don't clear the references
            # directly, the AtomGroup objects will still exist in the ExtendedAtom list even when they were already removed from an
            # instance of AtomGroupsManager().
            atm_grp.clear_refs()

            # Remove the atom group from the dict.
            key = tuple(sorted(atm_grp.atoms))
            if key in self.child_dict:
                del self.child_dict[key]

    def new_atm_grp(self, atoms, features=None, interactions=None):
        """Create a new `AtomGroup` object for ``atoms`` if one does not exist yet.
        Otherwise, return the existing `AtomGroup` object.

        Parameters
        ----------
        atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
            A sequence of atoms.
        features : iterable of :class:`~luna.mol.features.ChemicalFeature`, optional
            If provided, add ``features`` to a new or an already existing `AtomGroup` object.
        interactions : iterable of :class:`~luna.interaction.type.InteractionType`, optional
            If provided, add ``interactions`` to a new or an already existing `AtomGroup` object.

        Returns
        -------
         : AtomGroup
            A new or an already existing `AtomGroup` object.
        """
        key = tuple(sorted(atoms))
        features = features or []
        interactions = interactions or []

        if key in self.child_dict:
            atm_grp = self.child_dict[key]
        else:
            atm_grp = AtomGroup(atoms)
            self.add_atm_grps([atm_grp])

        if features:
            atm_grp.add_features(features)

        if interactions:
            atm_grp.add_interactions(interactions)

        return atm_grp

    def merge_hydrophobic_atoms(self, interactions_mngr):
        """Create hydrophobic islands by merging covalently bonded hydrophobic atoms in ``atm_grps``.
        Hydrophobic islands are atom groups having the feature `Hydrophobe`.

        Atom-atom hydrophobic interactions in ``interactions_mngr`` are also converted to
        island-island interactions.

        Parameters
        ----------
        interactions_mngr : :class:`~luna.interaction.calc.InteractionsManager`
            An :class:`~luna.interaction.calc.InteractionsManager` object from where
            hydrophobic interactions are selected and convert from atom-atom to island-island interactions.
        """

        # Only hydrophobic atom groups.
        hydrop_atm_grps = list(self.filter_by_types(["Hydrophobic"]))

        # Hydrophobic islands dictionary. Keys are integer values and items are defined by a set of atom groups.
        hydrop_islands = defaultdict(set)

        # It stores a mapping of an atom (represented by its full id) and a hydrophobic island (defined by its keys).
        atm_mapping = {}

        island_id = 0
        for atm_grp in hydrop_atm_grps:
            # Hydrophobic atoms are defined always as only one atom.
            atm = atm_grp.atoms[0]

            # Recover the groups of all neighbors of this atom (it will merge all existing islands).
            nb_grps = set([atm_mapping[nbi.full_id] for nbi in atm.neighbors_info if nbi.full_id in atm_mapping])

            # Already there are hydrophobic islands formed by the neighbors of this atom.
            if nb_grps:
                # Merge all groups of the neighbors of this atom.
                new_island = set(chain.from_iterable([hydrop_islands.pop(nb_grp_id) for nb_grp_id in nb_grps]))
                # Include this atom to the merged group.
                new_island.add(atm)

                for k in atm_mapping:
                    if atm_mapping[k] in nb_grps:
                        atm_mapping[k] = island_id

                hydrop_islands[island_id] = new_island
                atm_mapping[atm.get_full_id()] = island_id
            else:
                atm_mapping[atm.get_full_id()] = island_id
                hydrop_islands[island_id].add(atm)

            island_id += 1

        # Create AtomGroup objects for the hydrophobic islands
        for island_id in hydrop_islands:
            # It will update an existing atom group or create a new one with the informed parameters.
            hydrophobe = self.new_atm_grp(hydrop_islands[island_id], [ChemicalFeature("Hydrophobe")])
            # Update the island information
            hydrop_islands[island_id] = hydrophobe

        hydrop_interactions = list(interactions_mngr.filter_by_types(["Hydrophobic"]))
        island_island_inter = defaultdict(set)
        for inter in hydrop_interactions:
            src_atm = inter.src_grp.atoms[0]
            trgt_atm = inter.trgt_grp.atoms[0]

            # The two island ids are used as key.
            key = tuple(sorted([atm_mapping[src_atm.get_full_id()], atm_mapping[trgt_atm.get_full_id()]]))

            island_island_inter[key].add(inter)

        interactions = set()
        for k in island_island_inter:
            island_atms = defaultdict(set)
            for inter in island_island_inter[k]:
                src_atm = inter.src_grp.atoms[0]
                trgt_atm = inter.trgt_grp.atoms[0]

                island_atms[atm_mapping[src_atm.get_full_id()]].add(src_atm)
                island_atms[atm_mapping[trgt_atm.get_full_id()]].add(trgt_atm)

            centroid1 = im.centroid(im.atom_coordinates(island_atms[k[0]]))
            centroid2 = im.centroid(im.atom_coordinates(island_atms[k[1]]))
            cc_dist = im.euclidean_distance(centroid1, centroid2)

            params = {"dist_hydrop_inter": cc_dist}

            inter = InteractionType(hydrop_islands[k[0]], hydrop_islands[k[1]], "Hydrophobic",
                                    src_interacting_atms=island_atms[k[0]], trgt_interacting_atms=island_atms[k[1]], params=params)
            interactions.add(inter)

        # Update the list of interactions with the new island-island interactions.
        interactions_mngr.add_interactions(interactions)
        # Remove atom-atom hydrophobic interactions.
        interactions_mngr.remove_interactions(hydrop_interactions)

        for atm_grp in hydrop_atm_grps:
            features = [f for f in atm_grp.features if f.name != "Hydrophobic"]

            # It may happen that a atom group ends up having no feature after the remotion of the feature "Hydrophobic".
            # This is unlikely to occur as all atoms (by default) will have at least the feature 'Atom'. But, depending one the
            # pharmacophore rules definition, it can occur.
            atm_grp.features = features

    def get_shortest_path_length(self, src_grp, trgt_grp, cutoff=None):
        """Compute the shortest path length between two atom groups ``src_grp`` and ``trgt_grp``.

        The shortest path between two atom groups is defined as the shortest path between any of their atoms,
        which are calculated using Dijkstra’s algorithm and the graph ``graph``.

        If there is not any path between ``src_grp`` and ``trgt_grp``, infinite is returned.

        Parameters
        ----------
        src_grp, trgt_grp : `AtomGroup`
            Two atom groups to calculate the shortest path.
        cutoff : int
            Only paths of length <= ``cutoff`` are returned. If None, all path lengths are considered.

        Returns
        -------
         : int or float('inf'):
            The shortest path.
        """
        shortest_path_size = float('inf')
        for src_atm in src_grp.atoms:
            for trgt_atm in trgt_grp.atoms:
                try:
                    dist, path = single_source_dijkstra(self.graph, src_atm, trgt_atm, cutoff=cutoff)
                    if dist < shortest_path_size:
                        shortest_path_size = dist
                except Exception:
                    pass
        return shortest_path_size

    def save(self, output_file, compressed=True):
        """Write the pickled representation of the `AtomGroupsManager` object to the file ``output_file``.

        Parameters
        ----------
        output_file : str
            The output file.
        compressed : bool, optional
            If True (the default), compress the pickled representation as a gzip file (.gz).

        Raises
        -------
        FileNotCreated
            If the file could not be created.
        """
        pickle_data(self, output_file, compressed)

    @staticmethod
    def load(input_file):
        """Load the pickled representation of an `AtomGroupsManager` object saved at the file ``input_file``.

        Returns
        ----------
         : `AtomGroupsManager`
            The reconstituted `AtomGroupsManager` object, including its set of atom groups and interactions.

        Raises
        -------
        PKLNotReadError
            If the file could not be loaded.
        """
        return unpickle_data(input_file)

    def __len__(self):
        # Number of atom groups.
        return self.size

    def __iter__(self):
        """Iterate over children."""
        for atm_grp in self.atm_grps:
            yield atm_grp


class AtomGroup():
    """ Represent single atoms, chemical functional groups, or simply an arrangement of atoms as in hydrophobes.

    Parameters
    ----------
    atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
        A sequence of atoms.
    features : iterable of :class:`~luna.mol.features.ChemicalFeature`, optional
        A sequence of chemical features.
    interactions : iterable of :class:`~luna.interaction.type.InteractionType`, optional
        A sequence of interactions established by an atom group.
    recursive : bool
        If True, add the new atom group to the list of atom groups of each atom in ``atoms``.
    manager : `AtomGroupsManager`, optional
        The `AtomGroupsManager` object that contains this `AtomGroup` object.
    """

    def __init__(self, atoms, features=None, interactions=None, recursive=True, manager=None):
        self._atoms = sorted(atoms)

        # Atom properties
        self._coords = im.atom_coordinates(atoms)
        self._centroid = im.centroid(self.coords)
        self._normal = None

        features = features or []
        self._features = sorted(features)

        self._interactions = interactions or []
        self._hash_cache = None

        self._manager = manager

        self._recursive = recursive

        if recursive:
            for atm in self.atoms:
                atm.add_atm_grps([self])

    @property
    def atoms(self):
        """iterable of :class:`~luna.mol.atom.ExtendedAtom`, read-only: The sequence of atoms that belong to an atom group."""
        return self._atoms

    @property
    def compounds(self):
        """set of :class:`~luna.MyBio.PDB.Residue.Residue`, read-only: The set of unique compounds that contain the atoms in ``atoms``.

        As an atom group can be formed by the union of two or more compounds (e.g., amide of peptide bonds), it may return
        more than one compound.
        """
        return set([a.parent for a in self._atoms])

    @property
    def coords(self):
        """ array-like of floats : Atomic coordinates (x, y, z) of each atom in ``atoms``."""
        return self._coords

    @property
    def centroid(self):
        """ array-like of floats, read-only: The centroid (x, y, z) of the atom group.

        If ``atoms`` contains only one atom, then ``centroid`` returns the same as ``coords``.
        """
        return self._centroid

    @property
    def normal(self):
        """array-like of floats, read-only: The normal vector (x, y, z) of the points given by ``coords``."""
        if self._normal is None:
            self._normal = im.calc_normal(self.coords)
        return self._normal

    @property
    def features(self):
        """iterable of :class:`~luna.mol.features.ChemicalFeature`: A sequence of chemical features.

        To add or remove a feature use :py:meth:`add_features`
        or :py:meth:`remove_features`, respectively."""
        return self._features

    @features.setter
    def features(self, features):
        self._features = sorted(features)
        # Reset hash.
        self._hash_cache = None

    @property
    def feature_names(self):
        """iterable of str: The name of each chemical feature in ``features``."""
        return [f.name for f in self.features]

    @property
    def interactions(self):
        """iterable of :class:`~luna.interaction.type.InteractionType`: The sequence of interactions
        established by an atom group.

        To add or remove an interaction use :py:meth:`add_interactions`
        or :py:meth:`remove_interactions`, respectively."""
        return self._interactions

    @interactions.setter
    def interactions(self, interactions):
        self._interactions = interactions

    @property
    def manager(self):
        """`AtomGroupsManager`: The `AtomGroupsManager` object that contains an `AtomGroup` object."""
        return self._manager

    @manager.setter
    def manager(self, manager):
        if isinstance(manager, AtomGroupsManager):
            self._manager = manager
        else:
            raise IllegalArgumentError("The informed atom group manager must be an instance of '%s'." % AtomGroupsManager)

    @property
    def size(self):
        """int: The number of atoms comprising an atom group."""
        return len(self.atoms)

    def has_atom(self, atom):
        """Check if an atom group contains a given atom ``atom``.

        Parameters
        ----------
        atom : :class:`~luna.mol.atom.ExtendedAtom`

        Returns
        -------
         : bool
            If the atom group contains or not ``atom``.
        """
        return atom in self.atoms

    def contain_group(self, atm_grp):
        """Check if the atom group ``atm_grp`` is a subset of this atom group.

        For example, consider the benzene molecule.
        Its aromatic ring itself forms an `AtomGroup` object composed of all of its six atoms.
        Consider now any subset of carbons in the benzene molecule.
        This subset forms an `AtomGroup` object that is part of the group formed by the aromatic ring.
        Therefore, in this example, :meth:`contain_group` will return True because the aromatic ring
        contains the subset of hydrophobic atoms.

        Parameters
        ----------
        atm_grp : :class:`~luna.mol.groups.AtomGroup`

        Returns
        -------
         : bool
            If one atom group contains another atom group.
        """
        return set(atm_grp.atoms).issubset(set(self.atoms))

    def get_serial_numbers(self):
        """Get the serial number of each atom in an atom group."""
        return [a.get_serial_number() for a in self.atoms]

    def get_chains(self):
        """Get all unique chains in an atom group."""
        return sorted(set([a.get_parent_by_level("C").id for a in self.atoms]))

    def get_interactions_with(self, atm_grp):
        """Get all interactions that an atom group establishes with another atom group ``atm_grp``.

        Returns
        -------
         : iterable of :class:`~luna.interactions.type.InteractionType`
           All interactions established with the atom group ``atm_grp``.
        """
        target_interactions = []

        for inter in self.interactions:
            if inter.src_grp == atm_grp or inter.trgt_grp == atm_grp:
                target_interactions.append(inter)
        return target_interactions

    def get_shortest_path_length(self, trgt_grp, cutoff=None):
        """Compute the shortest path length between this atom group to another atom group ``trgt_grp``.

        The shortest path between two atom groups is defined as the shortest path between any of their atoms,
        which are calculated using Dijkstra’s algorithm.

        If ``manager`` is not provided, None is returned.

        If there is not any path between ``src_grp`` and ``trgt_grp``, infinite is returned.

        Parameters
        ----------
        trgt_grp : `AtomGroup`
            The target atom group to calculate the shortest path.
        cutoff : int, optional
            Only paths of length <= ``cutoff`` are returned. If None, all path lengths are considered.

        Returns
        -------
         : int, float('inf'), or None:
            The shortest path.
        """
        if self.manager is not None:
            return self.manager.get_shortest_path_length(self, trgt_grp, cutoff)
        return None

    def add_features(self, features):
        """ Add :class:`~luna.mol.features.ChemicalFeature` objects to ``features``."""
        self._features = sorted(set(self.features + list(features)))
        # Reset hash.
        self._hash_cache = None

    def remove_features(self, features):
        """ Remove :class:`~luna.mol.features.ChemicalFeature` objects from ``features``."""
        self._features = sorted(set(self.features) - set(features))
        # Reset hash.
        self._hash_cache = None

    def add_interactions(self, interactions):
        """ Add :class:`~luna.interaction.type.InteractionType` objects to ``interactions``."""
        self._interactions = list(set(self.interactions + list(interactions)))

    def remove_interactions(self, interactions):
        """ Remove :class:`~luna.interaction.type.InteractionType` objects from ``interactions``."""
        self._interactions = list(set(self.interactions) - set(interactions))

    def is_water(self):
        """Return True if all atoms in the atom group belong to water molecules."""
        return all([a.parent.is_water() for a in self.atoms])

    def is_hetatm(self):
        """Return True if all atoms in the atom group belong to hetero group, i.e., non-standard residues of proteins,
        DNAs, or RNAs, as well as atoms in other kinds of groups, such as carbohydrates,
        substrates, ligands, solvent, and metal ions.

        Hetero groups are designated by the flag HETATM in the PDB format."""
        return all([a.parent.is_hetatm() for a in self.atoms])

    def is_residue(self):
        """Return True if all atoms in the atom group belong to standard residues of proteins."""
        return all([a.parent.is_residue() for a in self.atoms])

    def is_nucleotide(self):
        """Return True if all atoms in the atom group belong to nucleotides."""
        return all([a.parent.is_nucleotide() for a in self.atoms])

    def is_mixed(self):
        """Return True if the atoms in the atom group belong to different compound classes
        (water, hetero group, residue, or nucleotide)."""
        return len(set([a.parent.get_class() for a in self.atoms])) > 1

    def has_water(self):
        """Return True if at least one atom in the atom group belongs to a water molecule."""
        return any([a.parent.is_water() for a in self.atoms])

    def has_hetatm(self):
        """Return True if at least one atom in the atom group belongs to a hetero group, i.e., non-standard residues of proteins,
        DNAs, or RNAs, as well as atoms in other kinds of groups, such as carbohydrates,
        substrates, ligands, solvent, and metal ions."""
        return any([a.parent.is_hetatm() for a in self.atoms])

    def has_residue(self):
        """Return True if at least one atom in the atom group belongs to a standard residue of proteins."""
        return any([a.parent.is_residue() for a in self.atoms])

    def has_nucleotide(self):
        """Return True if at least one atom in the atom group belongs to a nucleotide."""
        return any([a.parent.is_nucleotide() for a in self.atoms])

    def has_target(self):
        """Return True if at least one compound is the target of LUNA's analysis"""
        return any([a.parent.is_target() for a in self.atoms])

    def as_json(self):
        """Represent the atom group as a dict containing the atoms, compounds,
        features, and compound classes (water, hetero group, residue, or nucleotide).

        The dict is defined as follows:

            * ``atoms`` (iterable of :class:`~luna.mol.atom.ExtendedAtom`): the list of atoms comprising the atom group;
            * ``compounds`` (iterable of :class:`~luna.MyBio.PDB.Residue.Residue`): the list of unique compounds that contain the atoms;
            * ``classes`` (iterable of str): the list of compound classes;
            * ``features`` (iterable of :class:`~luna.mol.features.ChemicalFeature`): the atom group's list of chemical features.
        """
        grp_obj = {}
        grp_obj["atoms"] = [atm.as_json() for atm in self.atoms]
        grp_obj["compounds"] = [comp.as_json() for comp in self.compounds]
        grp_obj["features"] = [feat.name for feat in self.features if feat.name != "Atom"]
        grp_obj["classes"] = [comps.get_class() for comps in self.compounds]

        return grp_obj

    def clear_refs(self):
        """References to this `AtomGroup` instance will be removed from the list of atom groups of
        each atom in ``atoms``."""
        if self._recursive:
            for atm in self.atoms:
                atm.remove_atm_grps([self])

    def __repr__(self):
        return '<AtomGroup: [%s]>' % ', '.join([str(x) for x in self.atoms])

    def __eq__(self, other):
        """Overrides the default implementation"""
        if type(self) == type(other):
            return self.atoms == other.atoms and self.features == other.features
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __lt__(self, other):
        atms1 = tuple(sorted(self.atoms))
        atms2 = tuple(sorted(other.atoms))
        return atms1 < atms2

    def __len__(self):
        # Number of atoms.
        return self.size

    def __hash__(self):
        """Overrides the default implementation"""
        if self._hash_cache is None:
            # Transform atoms and features list into an imutable data structure.
            # The lists are sorted in order to avoid dependence on appending order.
            atoms_tuple = tuple(self.atoms)
            feat_tuple = tuple(self.features)
            self._hash_cache = hash((atoms_tuple, feat_tuple, self.__class__))
        return self._hash_cache


class PseudoAtomGroup(AtomGroup):
    """Represent only the atoms from an `AtomGroup` object that are involved in an interaction.

    Currently, this class is only used during the generation of LUNA's fingerprints.

    Parameters
    ----------
    parent_grp : `AtomGroup`
        The atom group that contains the subset of atoms ``atoms``.
    atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
        A sequence of atoms.
    features : iterable of :class:`~luna.mol.features.ChemicalFeature`, optional
        A sequence of chemical features.
    interactions : iterable of :class:`~luna.interaction.type.InteractionType`, optional
        A sequence of interactions established by an atom group.
    """

    def __init__(self, parent_grp, atoms, features=None, interactions=None):
        self.parent_grp = parent_grp

        super().__init__(atoms, features, interactions, recursive=False)

    def __repr__(self):
        return '<PseudoAtomGroup: [%s]>' % ', '.join([str(x) for x in self.atoms])


class AtomGroupPerceiver():

    """Perceive and create atom groups for molecules.

    Parameters
    ----------
    feature_extractor : :class:`~luna.mol.features.FeatureExtractor`
        Perceive pharmacophoric properties from molecules.
    add_h : bool
        If True, add hydrogen to the molecules.
    ph : float, optional
        If not None, add hydrogens appropriate for pH ``ph``.
    amend_mol : bool
        If True, apply validation and standardization of molecules read from a PDB file.
    charge_model : class:`~luna.mol.charge_model.ChargeModel`
        A charge model object. By default, the implementation of OpenEye charge model is used.
    mol_obj_type : {"rdkit", "openbabel"}
        If "rdkit", parse the converted molecule with RDKit and return an instance of :class:`rdkit.Chem.rdchem.Mol`.
        If "openbabel", parse the converted molecule with Open Babel and return an instance of
        :class:`openbabel.pybel.Molecule`.
    expand_selection : bool
        If True (the default), perceive features for a given molecule considering all nearby molecules.
        The goal is to identify any covalently bonded molecules that may alter the
        pharmacophoric properties or chemical functional groups.

        For instance, consider an amide of a peptide bond.
        If ``expand_selection`` is False, the residues forming the peptide bond will be analyzed separately,
        which will make the oxygen and the nitrogen of the amide to be perceived as carbonyl oxygen
        and amine, respectively.
        On the other hand, if ``expand_selection`` is True, the covalent bond between the residues will be
        identified and the amide will be correctly perceived.
    radius : float
        If ``expand_selection`` is True, select all molecules up to a maximum of ``radius`` away (measured in Å).
        The default value is 2.2, which comprises covalent bond distances.
    tmp_path : str, optional
        A temporary directory to where temporary files will be saved.
        If not provided, the system's default temporary directory will be used instead.
    critical : bool
        If False, ignore any errors during the processing a molecule and continue to the next one.
        The default value is True, which implies that any errors will raise an exception.

    Raises
    ------
    IllegalArgumentError
        If ``mol_obj_type`` is not either 'rdkit' nor 'openbabel'.
    """

    def __init__(self, feature_extractor, add_h=False, ph=None, amend_mol=True,
                 charge_model=OpenEyeModel(), mol_obj_type="rdkit",
                 expand_selection=True, radius=COV_SEARCH_RADIUS,
                 cache=None, tmp_path=None, critical=True):

        if mol_obj_type not in ACCEPTED_MOL_OBJ_TYPES:
            acc_mol_obj_types = ", ".join(ACCEPTED_MOL_OBJ_TYPES)
            raise IllegalArgumentError("Objects of type '%s' are not currently"
                                       "accepted. The available options are: "
                                       "%s." % (mol_obj_type,
                                                acc_mol_obj_types))

        self.feature_extractor = feature_extractor

        self.add_h = add_h
        self.ph = ph
        # If the user decided not to add hydrogens,
        # it will try to use the existing ones.
        self.keep_hydrog = not self.add_h

        self.amend_mol = amend_mol
        self.charge_model = charge_model
        self.mol_obj_type = mol_obj_type
        self.expand_selection = expand_selection
        self.radius = radius

        self.cache = cache
        self.tmp_path = tmp_path

        # If the pharmacophoric perception is critical, any exception during
        # the processing of a molecule will stop the whole processing.
        self.critical = critical

    def perceive_atom_groups(self, compounds, mol_objs_dict=None):
        """Perceive and create atom groups for each molecule in ``compounds``.

        Parameters
        ----------
        compounds : iterable of :class:`~luna.MyBio.PDB.Residue.Residue`
            A sequence of molecules.
        mol_objs_dict : dict
            Map a compound, represented by its id, to a molecular object (:class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`,
            or :class:`openbabel.pybel.Molecule`).

            This parameter can be used in cases where the ligand is read from a molecular file
            and no standardization or validation is required.

        Returns
        -------
         : `AtomGroupsManager`
            An `AtomGroupsManager` object containing all atom groups perceived for the molecules in ``compounds``.
        """

        mol_objs_dict = mol_objs_dict or {}

        self.atm_grps_mngr = AtomGroupsManager()
        self.atm_mapping = {}

        compounds = set(compounds)
        init_comps_set = set(compounds)

        if self.cache:
            cached_compounds = set([c for c in compounds
                                    if self.cache.is_compound_cached(c)])
            compounds = compounds - cached_compounds

        # Controls which compounds are allowed to expand
        # when 'expand_selection' is set ON.
        pruned_comps = set()
        # Create a queue of compounds to be processed.
        comp_queue = set(compounds)

        while comp_queue:
            try:
                comp = comp_queue.pop()

                logger.debug("It will try to perceive the features of the "
                             "compound '%s' as no predefined properties "
                             "were provided." % comp.full_name)

                mol_obj = mol_objs_dict.get(comp.id, None)

                # If no OBMol object was defined and expand_selection
                # was set ON and it is not a border compound, it will
                # get all compounds around the target and create a new
                # OBMol object with them.
                if mol_obj is None and self.expand_selection:
                    # Remove the ligand from the list when
                    # it was provided as an OBMol object.
                    comp_list = [c for c in get_proximal_compounds(comp)
                                 if c.id not in mol_objs_dict]

                    # Expands the queue of compounds
                    # with any new border compound.
                    if comp not in pruned_comps:
                        border_comps = set(comp_list) - compounds
                        pruned_comps |= border_comps
                        comp_queue |= border_comps

                # Otherwise, the compound list will be composed
                # only by the target compound.
                else:
                    comp_list = [comp]

                # Recover all atoms from the previous selected compounds.
                selector = Selector(keep_altloc=False,
                                    keep_hydrog=self.keep_hydrog)
                atoms = tuple([a for r in comp_list
                               for a in r.get_unpacked_list()
                               if selector.accept_atom(a)])

                self._assign_properties(comp, atoms, mol_obj)
            except Exception:
                logger.debug("Features for the compound '%s' were "
                             "not correctly perceived." % comp)

                if self.critical:
                    raise

        if self.cache and cached_compounds:
            self._apply_cache(cached_compounds, comp.get_parent_by_level("M"))

        # Remove atom groups not comprising the provided
        # compound list (parameter 'compounds').
        remove_atm_grps = []
        for atm_grp in self.atm_grps_mngr:
            if any([c in init_comps_set for c in atm_grp.compounds]) is False:
                remove_atm_grps.append(atm_grp)
        self.atm_grps_mngr.remove_atm_grps(remove_atm_grps)

        return self.atm_grps_mngr

    def _apply_cache(self, compounds, model):

        def copy_atom(ref_atm, force_update=False):
            ref_res = ref_atm.parent
            ref_chain = ref_res.parent

            # Recover current structure entities.
            chain = model[ref_chain.id]
            res = chain[ref_res.id]
            atm = res[ref_atm.id]

            # If an ExtendedAtom already exists, return it.
            if not force_update and atm in self.atm_mapping:
                return self.atm_mapping[atm]

            # Define a new ExtendedAtom.
            ext_atm = self._new_extended_atom(atm)

            # Set its atomic invariants.
            ext_atm.invariants = ref_atm.invariants

            # Add neighbors' information to the new ExtendedAtom.
            for nbi in ref_atm.neighbors_info:
                atom_info = AtomData(nbi.atomic_num,
                                     nbi.coord, nbi.bond_type,
                                     nbi.full_id)
                ext_atm.add_nb_info([atom_info])

            return ext_atm

        comp_names = set([c.full_name for c in compounds])

        for atm_grp in sorted(self.cache.atm_grps_mngr):
            if any([c.full_name in comp_names for c in atm_grp.compounds]):
                atoms = []
                for atm in atm_grp.atoms:
                    # Copy extended atom, i.e., create a new extended
                    # atom or recover an existing one and set invariants
                    # and neighbors' information when necessary.
                    ext_atm = copy_atom(atm, force_update=True)
                    atoms.append(ext_atm)

                # Create the new atom group.
                features = atm_grp.features
                self.atm_grps_mngr.new_atm_grp(atoms, features)

        for edge in sorted(self.cache.atm_grps_mngr.graph.edges):
            # Skip already existing edges.
            if edge in self.atm_grps_mngr.graph.edges:
                continue

            atoms = []
            for atm in edge:
                # Copy extended atom, i.e., create a new
                # extended atom or recover an existing one
                # and set invariants and neighbors' information
                # if necessary.
                ext_atm = copy_atom(atm)
                atoms.append(ext_atm)

            # Update the AtomGroupsManager object with a new edge.
            self.atm_grps_mngr.graph.add_edge(atoms[0], atoms[1], weight=1)

    def _assign_properties(self, target_compound, target_atoms, mol_obj=None):

        # If no OBMol was defined, create a new one with the compound list.
        ignored_atoms = []
        if mol_obj is None:
            mol_obj, ignored_atoms = self._get_mol_from_entity(target_compound, target_atoms)
        # Create a new MolWrapper object.
        mol_obj = MolWrapper(mol_obj)

        # If the add_h property is set to False, the code will not remove any existing hydrogens from the PDB structure.
        # In these situations, the list of atoms may contain hydrogens. But, we do not need to attribute properties to hydrogens.
        # We just need them to correctly set properties to heavy atoms. So let's just ignore them.
        target_atoms = [atm for atm in target_atoms if atm.element != "H" and atm not in ignored_atoms]

        if mol_obj.get_num_heavy_atoms() != len(target_atoms):
            raise MoleculeSizeError("The number of heavy atoms in the PDB selection and in the MOL file are different.")

        # Ignore hydrogen atoms.
        atm_obj_list = [atm for atm in mol_obj.get_atoms() if atm.get_atomic_num() != 1]

        atm_map = {}
        trgt_atms = {}
        for i, atm_obj in enumerate(atm_obj_list):
            atm_map[atm_obj.get_idx()] = target_atoms[i].get_full_id()

            trgt_atms[target_atoms[i].get_full_id()] = self._new_extended_atom(target_atoms[i])

            # Invariants are still None, so it means a new ExtendedAtom has just been created.
            if trgt_atms[target_atoms[i].get_full_id()].invariants is None:
                # Update atomic invariants for new ExtendedAtoms created.
                trgt_atms[target_atoms[i].get_full_id()].invariants = atm_obj.get_atomic_invariants()

            # The current atom already has its invariants updated, which means the atom was created previously
            # by another target compound.
            else:
                # In this case, if the current atom belongs to the target compound, we need to prioritize the current target compound
                # and overwrite the atom's invariants. By doing so, we always guarantee the most accurate information of an atom.
                #
                # It is done because some atoms are updated by centroids (target compound) not corresponding to its real compound.
                # When it happens, the information of covalently bonded atoms may be lost due to the short COV_SEARCH_RADIUS used when
                # selecting nearby compounds. As a consequence, these atoms will have their properties and topology incorrectly
                # perceived.
                #
                # However, we need to highlight that the main goal of selecting atoms using COV_SEARCH_RADIUS is to find atoms
                # covalently bonded to the current compound. Therefore, atoms in the border of nearby molecules are not important
                # right now and they will be updated in their own time.
                if trgt_atms[target_atoms[i].get_full_id()].parent == target_compound:
                    # Update the atomic invariants for an ExtendedAtom.
                    trgt_atms[target_atoms[i].get_full_id()].invariants = atm_obj.get_atomic_invariants()

        # Set all neighbors, i.e., covalently bonded atoms.
        for bond_obj in mol_obj.get_bonds():
            bgn_atm_obj = bond_obj.get_begin_atom()
            end_atm_obj = bond_obj.get_end_atom()

            # At least one of the atoms must be a non-hydrogen atom.
            if bgn_atm_obj.get_atomic_num() != 1 or end_atm_obj.get_atomic_num() != 1:
                # If the atom 1 is not a hydrogen, add atom 2 to its neighbor list.
                if bgn_atm_obj.get_atomic_num() != 1:
                    # If the current bgn_atm_obj consists of an atom from the current target compound, we can update its information.
                    # Other compounds are ignored for now as they will have their own time to update its information.
                    if trgt_atms[atm_map[bgn_atm_obj.get_idx()]].parent == target_compound:
                        full_id = atm_map.get(end_atm_obj.get_idx())
                        coord = mol_obj.get_atom_coord_by_id(end_atm_obj.get_id())
                        atom_info = AtomData(end_atm_obj.get_atomic_num(), coord, bond_obj.get_bond_type(), full_id)
                        trgt_atms[atm_map[bgn_atm_obj.get_idx()]].add_nb_info([atom_info])

                # If the atom 2 is not a hydrogen, add atom 1 to its neighbor list.
                if end_atm_obj.get_atomic_num() != 1:
                    # If the current end_atm_obj consists of an atom from the current target compound, we can update its information.
                    # Other compounds are ignored for now as they will have their own time to update its information.
                    if trgt_atms[atm_map[end_atm_obj.get_idx()]].parent == target_compound:
                        full_id = atm_map.get(bgn_atm_obj.get_idx())
                        coord = mol_obj.get_atom_coord_by_id(bgn_atm_obj.get_id())
                        atom_info = AtomData(bgn_atm_obj.get_atomic_num(), coord, bond_obj.get_bond_type(), full_id)
                        trgt_atms[atm_map[end_atm_obj.get_idx()]].add_nb_info([atom_info])

        group_features = self.feature_extractor.get_features_by_groups(mol_obj, atm_map)
        for key in group_features:
            grp_obj = group_features[key]

            # It only accepts groups containing at least one atom from the target compound.
            if any([trgt_atms[i].parent == target_compound for i in grp_obj["atm_ids"]]) is False:
                continue

            atoms = [trgt_atms[i] for i in grp_obj["atm_ids"]]
            self.atm_grps_mngr.new_atm_grp(atoms, grp_obj["features"])

        # Update the graph in the AtomGroupsManager object with the current compound information.
        for mybio_atm in target_compound.get_atoms():
            if mybio_atm.get_full_id() in trgt_atms and mybio_atm.parent == target_compound:
                atm = trgt_atms[mybio_atm.get_full_id()]

                for nb_info in atm.neighbors_info:
                    if nb_info.full_id in trgt_atms:
                        self.atm_grps_mngr.graph.add_edge(atm, trgt_atms[nb_info.full_id], weight=1)
        return True

    def _new_extended_atom(self, atm, invariants=None):
        if atm not in self.atm_mapping:
            self.atm_mapping[atm] = ExtendedAtom(atm, invariants=invariants)

        return self.atm_mapping[atm]

    def _get_mol_from_entity(self, target_compound, target_atoms):
        validate_mol = self.amend_mol
        standardize_mol = self.amend_mol and target_compound.is_residue()

        atom_selector = AtomSelector(target_atoms, keep_altloc=False, keep_hydrog=self.keep_hydrog)

        return biopython_entity_to_mol(target_compound.get_parent_by_level('M'), atom_selector,
                                       validate_mol=validate_mol, standardize_mol=standardize_mol, add_h=self.add_h,
                                       ph=self.ph, mol_obj_type=self.mol_obj_type, tmp_path=self.tmp_path)


class AtomGroupNeighborhood:
    """ Class for fast neighbor atom groups searching.

    ``AtomGroupNeighborhood`` makes use of a KD Tree implemented in C, so it's fast.

    Parameters
    ----------
    atm_grps : iterable of `AtomGroup`, optional
        A sequence of `AtomGroup` objects, which is used in the queries.
        It can contain atom groups from different molecules.
    bucket_size : int
        Bucket size of KD tree.
        You can play around with this to optimize speed if you feel like it.
        The default value is 10.

    """

    def __init__(self, atm_grps, bucket_size=10):
        self.atm_grps = list(atm_grps)

        # get the coordinates
        coord_list = [ga.centroid for ga in self.atm_grps]

        # to Nx3 array of type float
        self.coords = np.array(coord_list).astype("f")
        assert(bucket_size > 1)
        assert(self.coords.shape[1] == 3)
        self.kdt = KDTree(3, bucket_size)
        self.kdt.set_coords(self.coords)

    def search(self, center, radius):
        """Return all atom groups in ``atm_grps`` that is up to a maximum of ``radius`` away (measured in Å) of ``center``.

        For atom groups with more than one atom, their centroid is used as a reference.
        """
        self.kdt.search(center, radius)
        indices = self.kdt.get_indices()
        n_grps_list = []
        atm_grps = self.atm_grps
        for i in indices:
            a = atm_grps[i]
            n_grps_list.append(a)

        return n_grps_list



from os.path import basename, exists, isdir, isfile, splitext
from os import makedirs, remove, listdir
from shutil import rmtree
import string
import logging
import pickle
import gzip

from luna.util.exceptions import FileNotCreated, PKLNotReadError
from luna.util import new_random_string


logger = logging.getLogger()


def detect_compression_format(filename):
    """
    Attempts to detect compression format from the filename extension.
    Returns None if no format could be detected.
    """
    if filename.endswith('.bz2'):
        return "bz2"
    elif filename.endswith('.xz'):
        return "xz"
    elif filename.endswith('.gz'):
        return "gz"
    else:
        return None


def get_file_format(path, max_split=None, ignore_compression=False):
    """Detect the file format from pathname ``path``.

    Parameters
    ----------
    path : str
        The pathname.
    max_split : int or None
        Specifies how many splits to do. The default value is None, which is "all occurrences".
    ignore_compression: bool, optional
        Ignore compression format. The default value is False, which does not ignore compression format.

    Returns
    -------
    filename : str
        The file format.
    """
    if ignore_compression is True:
        comp_format = detect_compression_format(path)
        if comp_format:
            path = path[:-(len(comp_format) + 1)]
    return generic_splitext(path, max_split)[1][1:]


def get_filename(path, max_split=None):
    """Detect the filename from pathname ``path``.

    Parameters
    ----------
    path : str
        The pathname.
    max_split : int or None
        Specifies how many splits to do. The default value is None, which is "all occurrences".

    Returns
    -------
    filename : str
        The filename.
    """
    return generic_splitext(path, max_split)[0]


def generic_splitext(path, max_split=None):
    """Split the pathname ``path`` into a pair (``filename``, ``ext``).

    Parameters
    ----------
    path : str
        The pathname.
    max_split : int or None
        Specifies how many splits to do. The default value is None, which is "all occurrences".

    Returns
    -------
    filename : str
        The filename.
    ext: str
        The file extension.
    """
    filename = basename(path)

    num_ext = filename.count('.') + 1
    if max_split is None or max_split < 1:
        max_split = num_ext
    else:
        max_split = min(max_split, num_ext)

    filename, ext = splitext(filename)
    max_split -= 1
    while max_split > 0:
        filename, tmp_ext = splitext(filename)
        ext = tmp_ext + ext
        max_split -= 1

    return (filename, ext)


def generate_json_file(json_data, output_file, indent=4, sort_keys=True):
    """Serialize ``json_data`` to a JSON formatted string and save it at ``output_file``.

    Parameters
    ----------
    json_data : object
        The data to be serialized.
    output_file : str
        The output file where the serialized data will be saved.
    indent : int
        If indent is a non-negative integer or string, then JSON array elements and object members
        will be pretty-printed with that indent level. An indent level of 0, negative, or "" will only insert newlines.
        None selects the most compact representation. Using a positive integer indent indents that many spaces per level.
        If indent is a string (such as "\t"), that string is used to indent each level.
        The default value is 4.

    sort_keys : bool
        If ``sort_keys`` is True, the output of dictionaries will be sorted by key.
    """
    try:
        import simplejson as json
        logger.warning("Module 'simplejson' imported.")
    except ImportError:
        logger.warning("Module 'simplejson' not available. Built-in module 'json' will be imported.")
        import json

    try:
        with open(output_file, "w") as IN:
            json.dump(json_data, IN, indent=indent, sort_keys=sort_keys)
    except Exception as e:
        logger.exception(e)
        raise


def parse_json_file(json_file):
    """Deserialize the JSON file ``json_file``.

    Parameters
    ----------
    json_file : str
        The input JSON file.

    Returns
    -------
    json_data : object
        The deserialized data.
    """
    try:
        import simplejson as json
        logger.warning("Module 'simplejson' imported.")
    except ImportError:
        logger.warning("Module 'simplejson' not available. "
                       "Built-in module 'json' will be imported.")
        import json

    try:
        with open(json_file, "r") as IN:
            return json.load(IN)
    except Exception as e:
        logger.exception(e)
        raise


def create_directory(path, clear=False):
    """Create the directory pathaname ``path``.

    Parameters
    ----------
    path : str
        The directory pathname to be created.
    clear : bool
        If True, clear the directory if already exists.
        The default value is False.
    """
    try:
        if not exists(path):
            makedirs(path)
        elif clear:
            logger.info("The directory '%s' already exists, and it will be "
                        "cleared before the program continues." % path)
            remove_directory(path)
        else:
            logger.info("The directory '%s' already exists, but it will not "
                        "be cleared." % path)
    except OSError as e:
        logger.exception(e)
        raise


def remove_files(files):
    """Remove the provided files.

    Parameters
    ----------
    files : iterable
        An iterable object containing a sequence of files to be removed.
    """
    for f in files:
        if exists(f):
            remove(f)
        else:
            logger.info("File '%s' does not exist." % f)


def remove_directory(path, only_empty_paths=False):
    """Remove the directory given by the pathname ``path``.

    Parameters
    ----------
    path : str
        The pathname.
    only_empty_paths: bool, optional
        If True, do not remove non-empty directories. The default value is False, which removes all directories.
    """
    if isdir(path):
        # Do nothing if the directory is not empty and if only empty paths must be removed.
        if only_empty_paths and len(listdir(path)) != 0:
            return

        try:
            rmtree(path, ignore_errors=True)
        except OSError:
            raise


def new_unique_filename(path, size=32, chars=string.ascii_uppercase + string.digits, retries=5):
    """Generate a new unique random pathname.

    Parameters
    ----------
    path : str
        The target pathname.
    size : int, optional
        The size of the new filename. The default value is 32.
    chars : iterable, optional
        A sequence of characters to choose from. The default value is 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'.
    retries : int, optional
        The function will keep trying to generate a unique filename (not exist in the pathname ``path``)
        until the maximum number of retries ``retries`` is reached.

    Returns
    -------
    unique_pathname : str
        A new random unique pathname (path + random filename).
    """
    for r in range(retries):
        filename = '%s/%s' % (path, new_random_string(size, chars))
        if not exists(filename):
            return filename


def validate_filesystem(path, type):
    """Validate if a file or directory exists and if it has the expected type.

    Parameters
    ----------
    path : str
        The pathname of the file or directory to be validated.
    type : {'file', 'directory'}
        The filesystem type.

    Raises
    -------
    FileNotFoundError
        If the file or directory given by the pathname ``path`` does not exist.
    IsADirectoryError
        If a file is provided but a directory is found instead at the pathname ``path``.
    NotADirectoryError
        If the filesystem given by the pathname ``path`` is not a directory.
    """
    if exists(path) is False:
        raise FileNotFoundError("File or directory '%s' does not exist" % path)

    if type == "file" and isfile(path) is False:
        raise IsADirectoryError("'%s' is not a file." % path)
    elif type == "directory" and isdir(path) is False:
        raise NotADirectoryError("'%s' is not a directory." % path)


def is_directory_valid(path):
    """Check if ``path`` exists and if it is in fact a directory."""
    try:
        validate_filesystem(path, "directory")
        return True
    except Exception:
        return False


def is_file_valid(path):
    """Check if ``path`` exists and if it is in fact a file."""
    try:
        validate_filesystem(path, "file")
        return True
    except Exception:
        return False


def validate_directory(path):
    """Validate ``path`` as a directory."""
    validate_filesystem(path, "directory")


def validate_file(path):
    """Validate ``path`` as a file."""
    validate_filesystem(path, "file")


def pickle_data(data, output_file, compressed=True):
    """Write the pickled representation of the object ``data`` to the file ``output_file``.

    Parameters
    ----------
    data : object
        The object to be pickled.
    output_file : str
        The output file where the pickled representation will be saved.
    compressed : bool, optional
        If True (the default), compress the pickled representation as a gzip file (.gz).

    Raises
    -------
    FileNotCreated
        If the file could not be created.
    """
    open_func = open
    if compressed:
        open_func = gzip.open
        if output_file.endswith(".gz") is False:
            output_file += ".gz"

    try:
        with open_func(output_file, "wb") as OUT:
            pickle.dump(data, OUT, pickle.HIGHEST_PROTOCOL)
    except OSError as e:
        raise FileNotCreated("File '%s' could not be created." % output_file) from e
    except Exception:
        raise


def unpickle_data(input_file):
    """Read the pickled representation of an object from the file ``input_file`` and
    return the reconstituted object hierarchy specified therein.
    ``input_file`` can be a gzip-compressed file.

    Raises
    -------
    PKLNotReadError
        If the file could not be loaded.
    """
    try:
        # Try to decompress and unpickle the data first.
        with gzip.open(input_file, "rb") as IN:
            return pickle.load(IN)
    except Exception:
        pass

    # If the decompression failed, let's try to unpickle the file directly. Maybe it is not a compressed file.
    try:
        with open(input_file, "rb") as IN:
            return pickle.load(IN)
    except OSError as e:
        raise PKLNotReadError("File '%s' could not be loaded." % input_file) from e
